# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import json
from pathlib import Path
from typing import Any

from ai4rag import logger

from .base_event_handler import BaseEventHandler, EvaluationRecord, LogLevel, PatternPayload


class LocalEventHandler(BaseEventHandler):
    """Implementation used for local development and testing purposes.

    Parameters
    ----------
    output_path : str | Path | None, default=None
        Location where results for all patterns will be saved.
    """

    def __init__(self, output_path: str | Path | None = None):
        self.output_path = Path(output_path) if output_path else None

    def on_status_change(self, level: LogLevel, message: str, step: str | None = None) -> None:
        logger.info("LocalEventHandler ::: %s ::: %s ::: %s", level, step, message)

    def on_pattern_creation(
        self, payload: PatternPayload, evaluation_results: list[EvaluationRecord], **kwargs
    ) -> None:
        logger.info("LocalEventHandler ::: Pattern creation ::: %s", payload)
        pattern_name = payload.get("pattern_name", "default_pattern_name")

        if self.output_path:
            dir_path = self.output_path / pattern_name
            dir_path.mkdir(exist_ok=False, parents=True)

            evaluation_results_path = dir_path / "evaluation_results.json"
            with open(evaluation_results_path, encoding="utf-8", mode="w") as file:
                json.dump(evaluation_results, file)

            with open(dir_path / "pattern.json", encoding="utf-8", mode="w") as file2:
                json.dump(payload, file2)


class KFPEventHandler(BaseEventHandler):
    """Event handler that aggregates status changes and created patterns for post-experiment use.

    All status changes are collected in :attr:`status_changes` and all pattern results in
    :attr:`patterns`. When ``step`` is omitted from :meth:`on_status_change`, the last known
    step value is reused so every entry always carries a step label.

    To be used within kubeflow-pipelines components.
    """

    def __init__(self):
        self.status_changes: list[dict] = []
        self.patterns: list[dict[str, Any]] = []
        self._last_step: str | None = None

    def on_status_change(self, level: LogLevel, message: str, step: str | None = None) -> None:
        if step is not None:
            self._last_step = step
        self.status_changes.append({"level": level, "message": message, "step": self._last_step})

    def on_pattern_creation(
        self, payload: PatternPayload, evaluation_results: list[EvaluationRecord], **kwargs
    ) -> None:
        self.patterns.append({"payload": payload, "evaluation_results": evaluation_results, **kwargs})



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Literal

from pydantic_core import ErrorDetails

from ai4rag.search_space.src.exceptions import SearchSpaceValueError
from ai4rag.utils.constants import SearchSpaceValidationErrors

__all__ = ["validation_error_decoder"]


def _get_error_constrain_name_with_location(
    location: tuple[int | str, ...], skip_last_element: bool = True
) -> tuple[str, int]:
    """
    Retrieves exact object location where the ValidationError occured.

    Parameters
    ----------

    location : tuple[int | str, ...]
        Location of the ValidationError.

    skip_last_element : bool = True
        Set to True when we want to omit last location element.

    Returns
    -------
    tuple[str, int]
        Tuple with constrain attribute and position in the list if the parameter is a list.

    Example
    -------
    ```python
    location = ('chunking', 0, 'chunk_size')

    _get_error_constrain_name_with_location(location, skip_last_element=False)

    >>> (chunking.chunk_size, 0)
    ```
    """

    position = list((filter(lambda x: isinstance(x, int), location)))

    if not position:
        return str(location[0]), 0

    position_index = location.index(position[0])
    if skip_last_element:
        location_slice = slice(position_index + 1, -1)
    else:
        location_slice = slice(position_index + 1, None)

    message_constraints = location[:position_index] + location[location_slice]
    constrain_name = ".".join(message_constraints)
    return constrain_name, int(position[0])


def _get_parameter_length(
    validation_error: ErrorDetails, parameter_name: Literal["min_length", "max_length"]
) -> tuple[str, str]:
    """
    Retrieves actual parameter length from validation error and valid parameter length.

    Parameters
    ----------

    validation_error : ErrorDetails
        Dict like object which stores error information.

    parameter_name : Literal["min_length", "max_length"]
        Parameter name for which parameter length is retrieved.

    Returns
    -------
    tuple[str, str]
        Tuple which contains required min or max length of the parameter and the
        actual input length which caused the error.

    Example
    -------
    ```python
    error_details: ErrorDetails = {
        "type": "too_long",
        "loc": ("chunking",),
        "msg": "List should have at most 4 items after validation, not 5",
        "input": [
            {"chunk_size": 1000},
            {"chunk_size": 900},
            {"chunk_size": 800},
            {"chunk_size": 700},
            {"chunk_size": 600},
        ],
        "ctx": {"field_type": "List", "max_length": 4, "actual_length": 5},
        "url": "https://errors.pydantic.dev/2.9/v/too_long",
    }
    _get_parameter_length(error_details, "max_length")
    >>> (4, 5)
    ```
    """
    context = validation_error.get("ctx", {})
    required_items_num = f"{items} item" if (items := context.get(parameter_name)) == 1 else f"{items} items"
    input_items_num = context.get("actual_length", "")
    return required_items_num, input_items_num


def _get_parameter_limes_value_and_input(
    validation_error: ErrorDetails, parameter_name: Literal["le", "ge"]
) -> tuple[str, str]:
    """
    Retrieves parameter limes (upper or lower bound) and input which caused the validation error.

    Parameters
    ----------

    validation_error : ErrorDetails
        Dict like object which stores error information.

    parameter_name : Literal["le", "ge"]
        Parameter name for which parameter length is retrieved.

    Returns
    -------
    tuple[str, str]
        Tuple with upper or lower bound for the parameter value and the
        input value which caused the error.

    Example
    -------
    error_details: ErrorDetails = {
        'type': 'less_than_equal',
        'loc': ('chunking', 0, 'chunk_size'),
        'msg': 'Input should be less than or equal to 1024',
        'input': 50000, 'ctx': {'le': 1024},
        'url': 'https://errors.pydantic.dev/2.9/v/less_than_equal'
    }
    _get_parameter_limes_value_and_input(error_details, "le")
    >>> (1024, 50000)
    """
    context = validation_error.get("ctx", {})
    limes = str(context.get(parameter_name, 0))
    input_value = validation_error.get("input")
    return limes, input_value


def validation_error_decoder(
    validation_error: ErrorDetails,
) -> None:
    """
    Tiny wrapper for Pydantic ValidationError.
    Unpacks aforementioned error and raise it as SearchSpaceValueError.
    Every SearchSpaceValueError is re-raised from None to get rid of Pydantic traceback.

    Parameters
    ----------

    validation_error : ErrorDetails
        Dict like object which stores error information.

    Raises
    ------

    SearchSpaceValueError
        Context of this error depends on which ValidationError type occured during payload validation.

    Examples
    --------
    First example:

    ```python
    error_details: ErrorDetails = {
        'type': 'list_type', 'loc': ('chunking',),
        'msg': 'Input should be a valid list',
        'input': 'invalid_key',
        'url': 'https://errors.pydantic.dev/2.9/v/list_type'
    }
    validation_error_decoder(error_details) # match for SearchSpaceValidationErrors.LIST_TYPE
    ```

    Second example:
    ```python
    error_details: ErrorDetails = {
        'type': 'less_than_equal',
        'loc': ('chunking', 0, 'chunk_size'),
        'msg': 'Input should be less than or equal to 1024',
        'input': 50000, 'ctx': {'le': 1024},
        'url': 'https://errors.pydantic.dev/2.9/v/less_than_equal'
    }
    validation_error_decoder(error_details) # match for SearchSpaceValidationErrors.LESS_THAN_EQUAL
    ```
    """

    # if SearchSpaceValueError occurred during validation then it will be re-raised here
    if (context := validation_error.get("ctx")) and (error := context.get("error")):
        raise error from None

    location = validation_error.get("loc")

    match validation_error.get("type"):
        case SearchSpaceValidationErrors.UNEXPECTED_KEYWORD_ARGUMENT:

            if len(location) == 1:
                raise SearchSpaceValueError(f"{location[0]} is not a recognized parameter.") from None

            constrain_name, position = _get_error_constrain_name_with_location(location=location)
            raise SearchSpaceValueError(
                f"Constraint `{constrain_name}` at position {position} got invalid parameter name `{location[-1]}`."
            ) from None

        case SearchSpaceValidationErrors.LITERAL_ERROR:
            constrain_name, position = _get_error_constrain_name_with_location(
                location=location, skip_last_element=False
            )
            invalid_literal_value = validation_error.get("input")
            valid_literal_values = validation_error.get("ctx", {}).get("expected")
            raise SearchSpaceValueError(
                f"Constraint `{constrain_name}` at position {position} got invalid value `{invalid_literal_value}`, "
                f"but expected value is {valid_literal_values}."
            ) from None

        case SearchSpaceValidationErrors.TOO_SHORT:
            constrain_name = location[0]
            required_items_num, input_items_num = _get_parameter_length(
                validation_error=validation_error, parameter_name="min_length"
            )
            raise SearchSpaceValueError(
                f"Constraint `{constrain_name}` should have at least {required_items_num}, not `{input_items_num}`."
            ) from None

        case SearchSpaceValidationErrors.TOO_LONG:
            constrain_name = location[0]
            required_items_num, input_items_num = _get_parameter_length(
                validation_error=validation_error, parameter_name="max_length"
            )
            raise SearchSpaceValueError(
                f"Constraint `{constrain_name}` should have at most {required_items_num}, not `{input_items_num}`."
            ) from None

        case SearchSpaceValidationErrors.LESS_THAN_EQUAL:
            constrain_name, position = _get_error_constrain_name_with_location(
                location=location, skip_last_element=False
            )
            required_num_bound, input_num = _get_parameter_limes_value_and_input(
                validation_error=validation_error, parameter_name="le"
            )
            raise SearchSpaceValueError(
                f"Constraint `{constrain_name}` at position {position} should be less than or equal to "
                f"{required_num_bound}, but is equal to `{input_num}`."
            ) from None

        case SearchSpaceValidationErrors.GREATER_THAN_EQUAL:
            constrain_name, position = _get_error_constrain_name_with_location(
                location=location, skip_last_element=False
            )
            required_num_bound, input_num = _get_parameter_limes_value_and_input(
                validation_error=validation_error, parameter_name="ge"
            )
            raise SearchSpaceValueError(
                f"Constraint `{constrain_name}` at position {position} should be higher than or equal to "
                f"{required_num_bound}, but is equal to `{input_num}`."
            ) from None

        case (
            SearchSpaceValidationErrors.INT_FROM_FLOAT
            | SearchSpaceValidationErrors.INT_PARSING
            | SearchSpaceValidationErrors.INT_TYPE
            | SearchSpaceValidationErrors.LIST_TYPE
        ):
            constrain_name = ".".join([str(location_piece) for location_piece in location])
            constrain_input = validation_error.get("input")
            raise SearchSpaceValueError(
                f"Constraint {constrain_name} got invalid value {constrain_input}. "
                f"Check the docs for valid input constraints."
            ) from None

        case _:
            # it should not be raised in any case
            # if you will trigger the below error then wrap it up as shown above
            constrain_name = ".".join([str(location_piece) for location_piece in location])
            constrain_input = validation_error.get("input")
            raise SearchSpaceValueError(
                f"Unknown validation error occurred for constraint `{constrain_name}` for input {constrain_input}."
            ) from None



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Annotated, Optional

from annotated_types import MinLen
from pydantic import (
    BaseModel,
    ConfigDict,
)

CONFIG = ConfigDict(extra="forbid")


class AI4RAGFoundationModel(BaseModel):
    """Attributes to be included in the generation.foundation_models payload."""

    model_config = CONFIG

    model_id: Annotated[str, MinLen(1)]


class AI4RAGEmbeddingModel(BaseModel):
    """Attributes to be included in the generation.embedding_models payload."""

    model_config = CONFIG

    model_id: Annotated[str, MinLen(1)]


class AI4RAGConstraints(BaseModel):
    """Attributes to be included in constraints payload."""

    model_config = CONFIG

    embedding_models: Optional[Annotated[list[AI4RAGEmbeddingModel], MinLen(1)]] = None
    foundation_models: Optional[Annotated[list[AI4RAGFoundationModel], MinLen(1)]] = None



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import random
from copy import copy
from dataclasses import dataclass
from math import ceil
from typing import Any, Callable

import numpy as np
import pandas as pd
from pygam import LinearGAM
from sklearn.preprocessing import LabelEncoder

from ai4rag import logger
from ai4rag.core.hpo.base_optimizer import BaseOptimizer, FailedIterationError, OptimizationError, OptimizerSettings
from ai4rag.search_space.src.search_space import SearchSpace

__all__ = ["GAMOptSettings", "GAMOptimizer"]


@dataclass
class GAMOptSettings(OptimizerSettings):
    """
    Settings for the GAMOptimizer. For the detailed description
    of parameters for Generalized Additive Models, please see pygam
    documentation.
    """

    n_random_nodes: int = 4
    evals_per_trial: int = 1
    random_state: int = 64


class GAMOptimizer(BaseOptimizer):
    """
    Optimizer based on Generalized Additive Models.
    Trained model is used to suggest next node in the search space
    for evaluation.

    Parameters
    ----------
    objective_function : Callable[[dict], float]
        Target function that will be used in every evaluation. Output of
        this function should be 'float', as this is the value for which algorithms
        try to optimize solution. Function should take dict filled with 'key: value' pairs
        that are 'argument: corresponding value'.

    search_space : SearchSpace
        Instance containing information about nodes in the solutions space that
        will be evaluated during the optimization.

    settings : GAMOptSettings
        Instance with settings required for configuring the optimization process.

    Attributes
    ----------
    evaluations : list[dict]
        Already evaluated hyperparameters combinations with corresponding score.

    max_iterations : int
        Validated maximum number of iterations during HPO.
    """

    def __init__(
        self,
        objective_function: Callable[[dict], float],
        search_space: SearchSpace,
        settings: GAMOptSettings,
        known_observations: list[dict] | None = None,
    ):
        super().__init__(objective_function, search_space, settings)
        self.settings = settings
        self.evaluations = []
        self._evaluated_combinations = []
        self._encoders_with_columns: list[tuple[str, LabelEncoder]] = []

        if known_observations:
            self._load_known_observations(known_observations)

        self.max_iterations = self.settings.max_evals

    @property
    def max_iterations(self) -> int:
        """Get max possible number of iterations for the HPO."""
        return self._max_iterations

    @max_iterations.setter
    def max_iterations(self, val: int) -> None:
        """Set maximum number of iterations that should be performed during HPO."""
        max_comb = self._search_space.max_combinations
        if val > max_comb:
            logger.info(
                (
                    "'max_number_of_rag_patterns' exceeded number of possible combinations: %s. "
                    "Setting 'max_number_of_rag_patterns' to: %s"
                ),
                max_comb,
                max_comb,
            )
            self._max_iterations = max_comb
        else:
            self._max_iterations = val

    def search(self) -> dict[str, Any]:
        """
        Actual function performing hyperparameter optimization for the selected
        objective function.

        Returns
        -------
        dict[str, Any]
            The best set of parameters with achieved score.

        Raises
        ------
        OptimizationError
            When there were no successful evaluations for given constraints.
        """
        self.evaluate_initial_random_nodes()

        iterations_limit = self._get_iterations_limit()

        for _ in range(iterations_limit):
            self._run_iteration()

        successful_evaluations = [evaluation for evaluation in self.evaluations if evaluation["score"] is not None]
        if not successful_evaluations:
            raise OptimizationError("Number of evaluations has reached limit. All iterations have failed.")

        # Sort in ascending order and take the last element (highest score).
        # This assumes we're maximizing the score.
        best_config_with_score = sorted(successful_evaluations, key=lambda d: d["score"])[-1]

        return best_config_with_score

    def _get_iterations_limit(self) -> int:
        """
        Calculate maximum number of iterations that can be proceeded based on the
        already evaluated random nodes and settings for the optimizer.
        """
        iterations_limit = ceil((self.max_iterations - len(self.evaluations)) / self.settings.evals_per_trial)
        return iterations_limit

    def _load_known_observations(self, known_observations: list[dict]) -> None:
        """
        Load known observations to warm-start the optimizer.

        Parameters
        ----------
        known_observations : list[dict]
            List of previously evaluated parameter combinations with scores.
            Each dict must contain the same keys as search space combinations
            plus a "score" key.

        Raises
        ------
        ValueError
            When any observation is missing the "score" key.
        """
        for idx, obs in enumerate(known_observations):
            if "score" not in obs:
                raise ValueError(f"Known observation at index {idx} is missing the 'score' key.")

            params = {k: v for k, v in obs.items() if k != "score"}
            self._evaluated_combinations.append(params)
            self.evaluations.append(obs.copy())

        logger.info("Loaded %d known observations into the optimizer.", len(known_observations))

    def evaluate_initial_random_nodes(self) -> None:
        """
        Perform evaluation of randomly chosen n nodes from the solutions space.
        Evaluations are performed until desired number of successful evaluations
        is reached or maximum number of evaluations is reached.

        When the optimizer has been warm-started with known observations,
        already-successful evaluations count toward the n_random_nodes target
        and already-evaluated combinations are excluded from candidates.
        """
        successful_evaluations = sum(1 for e in self.evaluations if e["score"] is not None)

        if successful_evaluations >= self.settings.n_random_nodes:
            logger.info(
                "Skipping random evaluation phase: %d known successful evaluations >= n_random_nodes (%d).",
                successful_evaluations,
                self.settings.n_random_nodes,
            )
            return

        if len(self.evaluations) >= self.max_iterations:
            return

        combinations_local = [c for c in copy(self._search_space.combinations) if c not in self._evaluated_combinations]
        random.shuffle(combinations_local)

        gen = (x for x in combinations_local)

        while successful_evaluations < self.settings.n_random_nodes:
            params = next(gen)
            score = self._objective_function(params=params)
            if score is not None:
                successful_evaluations += 1
            self._evaluated_combinations.append(params)
            params_with_score = params | {"score": score}
            self.evaluations.append(params_with_score)

            if len(self.evaluations) == self.max_iterations:
                break

    # pylint: disable=too-many-locals
    def _run_iteration(self) -> None:
        """
        Run single optimization iteration that consists of training GAM model
        to predict score for remaining nodes in the solutions space and choose
        the best n ones for further evaluation.
        """
        self._prepare_encoder()
        df = pd.DataFrame(data=self.evaluations)  # --> These are already known observations with scores.
        df = df[df["score"].notna()].copy()
        data = df.drop(columns=["score"])
        target = df["score"]

        x_train_enc = []
        for column, encoder in self._encoders_with_columns:
            x_train_enc.append(encoder.transform(data[column]))
        x_train_enc = np.column_stack(x_train_enc)

        gam = LinearGAM()
        gam.fit(x_train_enc, target)

        remaining_evaluations = self._get_remaining_evaluations(
            self._search_space.combinations, self._evaluated_combinations
        )

        remaining_evaluations_df = pd.DataFrame(remaining_evaluations)

        # Optimize encoding: build array directly
        encoded_data_to_predict = np.column_stack(
            [encoder.transform(remaining_evaluations_df[column]) for column, encoder in self._encoders_with_columns]
        )

        predictions = gam.predict(encoded_data_to_predict)

        for idx, val in enumerate(remaining_evaluations):
            val["score"] = predictions[idx]

        # Sort in descending order to get highest predictions first
        best_predictions = sorted(remaining_evaluations, key=lambda d: d["score"], reverse=True)

        n_best_predictions = best_predictions[: self.settings.evals_per_trial]

        for params in n_best_predictions:
            params.pop("score", None)
            score = self._objective_function(params)
            self._evaluated_combinations.append(params)
            self.evaluations.append(params | {"score": score})

    def _prepare_encoder(self) -> None:
        """
        Prepare encoder for the further processing based on all available combinations.
        """
        if not self._encoders_with_columns:
            logger.debug("Preparing encoder for %s...", self.__class__.__name__)
            df = pd.DataFrame(data=self._search_space.combinations)
            for column in df.columns:
                self._encoders_with_columns.append((column, LabelEncoder().fit(df[column])))
            logger.debug("Encoder for %s has been prepared.", self.__class__.__name__)

    @staticmethod
    def _get_remaining_evaluations(all_combinations: list[dict], evaluations: list[dict]) -> list[dict]:
        """
        Get all evaluations that has not been yet proceeded.

        Parameters
        ----------
        all_combinations : list[dict]
            All possible combinations of parameters.

        evaluations : list[dict]
            Combinations that have already been evaluated.

        Returns
        -------
        list[dict]
            Remaining combinations that have not yet been evaluated.
        """
        remaining = []

        for ev in all_combinations:
            if ev not in evaluations:
                remaining.append(ev.copy())

        return remaining

    # pylint: disable=duplicate-code
    def _objective_function(self, params: dict) -> float | None:
        """
        Wrapper around the objective function provided to the optimizer.

        Parameters
        ----------
        params : dict
            A dictionary containing parameters of pattern to be evaluated.

        Returns
        -------
        float | None
            Optimization score achieved for single node evaluation.
            If None - iteration has ended up with a failed status.
        """

        try:
            logger.info("Evaluating objective function with parameters: %s", params)
            loss = self.objective_function(params)

        except FailedIterationError:
            # None is here to avoid penalization of iterations failing due to unknown reasons
            loss = None

        return loss



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2025-2026.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import hashlib
from datetime import datetime
from typing import Any, cast

from langchain_chroma import Chroma
from langchain_core.documents import Document

from ai4rag import logger

from ..embedding.base_model import BaseEmbeddingModel
from .base_vector_store import BaseVectorStore
from .utils import merge_window_into_a_document


class ChromaVectorStore(BaseVectorStore):
    """
    Class representing single index in the chroma vector database.

    Parameters
    ----------
    embedding_model : BaseEmbeddingModel
        Instance used for embedding documents and user's queries.

    reuse_collection_name : str, default=None
        Name of the collection that will be created as a vector store.

    distance_metric : str, default="cosine"
        Metric that will be used to calculate similarity score between vectors.

    document_name_field : str, default="document_id"
        Default document ID field name.

    chunk_sequence_number_field : str, default="chunk_sequence_number"
        Default chunk sequence number field name.
    """

    _supported_distance_metrics = ("cosine", "l2")

    def __init__(
        self,
        embedding_model: BaseEmbeddingModel,
        reuse_collection_name: str | None = None,
        distance_metric: str = "cosine",
        document_name_field: str = "document_id",
        chunk_sequence_number_field: str = "sequence_number",
        **kwargs,
    ) -> None:
        super().__init__(
            embedding_model=embedding_model,
            distance_metric=distance_metric,
            reuse_collection_name=reuse_collection_name,
        )
        self._document_name_field = document_name_field
        self._chunk_sequence_number_field = chunk_sequence_number_field
        self._collection_name = reuse_collection_name or kwargs.pop(
            "collection_name", f"ai4rag_{datetime.now().strftime("%Y%m%d%H%M%S")}"
        )
        self._vector_store = self._get_chroma_client(**kwargs)

    def _get_chroma_client(self, **kwargs) -> Chroma:
        """
        Create chroma client based on the given settings.

        ^kwargs are passed from the __init__ as parameters for Chroma client.

        Returns
        -------
        Chroma
            Client instance created based on the given settings.
        """

        chroma_client = Chroma(
            collection_name=self.collection_name,
            embedding_function=self.embedding_model,
            collection_metadata={"hnsw:space": self.distance_metric},
            **kwargs,
        )

        return chroma_client

    @property
    def distance_metric(self) -> str:
        """Get used distance metric."""
        return self._distance_metric

    @distance_metric.setter
    def distance_metric(self, value: str) -> None:
        """Set value of the distance metric.

        Raises
        ------
        ValueError
            If the distance metric is not supported.
        """
        if value not in self._supported_distance_metrics:
            raise ValueError(f"Invalid distance metric: {value}. Use one of: {self._supported_distance_metrics}.")
        self._distance_metric = value

    @property
    def collection_name(self) -> str:
        """Dynamically get collection name."""
        return self._collection_name

    def clear(self) -> None:
        """Clear the vector store."""
        all_docs_ids = self._vector_store.get()["ids"]
        if len(all_docs_ids) > 0:
            self.delete(all_docs_ids)

    def count(self) -> int:
        """Count the number of shards in the vector store.

        Returns
        -------
        int
            Number of shards in the vector store.
        """
        return len(self._vector_store.get()["ids"])

    @staticmethod
    def _as_langchain_documents(content: list) -> list[Document]:
        """Creates a LangChain ``Document`` list from a list of potentially unstructured data.

        Parameters
        ----------
        content : list
            Unstructured data to be parsed.

        Returns
        -------
        list[Document]
            List of `Document` instances.
        """
        result = []
        for doc in content:
            if isinstance(doc, str):
                result.append(Document(page_content=doc))
            elif isinstance(doc, dict):
                content_str: str | None = doc.get("content", None)
                metadata = doc.get("metadata", {})

                if content_str:
                    if isinstance(metadata, dict):
                        result.append(Document(page_content=content_str, metadata=metadata))
                    else:
                        logger.warning(
                            "Document: %s is incorrect. Metadata needs to be given with 'metadata' "
                            "attribute and it needs to be a serializable dict. Skipping.",
                            doc,
                        )
                        continue
                else:
                    logger.warning("Document: %s is incorrect. Field 'content' is required.", doc)
                    continue
            else:
                try:
                    result.append(Document(page_content=doc.page_content, metadata=doc.metadata))
                except AttributeError:
                    logger.warning(
                        "Document: %s is not a dict, nor string, nor LangChain Document-like object. Skipping.", doc
                    )

        return result

    def _process_documents(self, content: list) -> tuple[list[str], list[Document]]:
        """
        Processes arbitrary list of data to produce two lists:
        one with unique IDs, and one with LangChain documents.

        Handles duplicate documents.

        Parameters
        ----------
        content : list
            Arbitrary data.

        Returns
        -------
        tuple[list[str], list[Document]]
            Lists with IDs and docs
        """
        docs = self._as_langchain_documents(content)
        if docs:
            # Take only unique ID document. Get two lists, one with ids, one with documents
            # For some documents, not all chars can be encoded properly.
            # In such cases, replace invalid chars by question marks, i.e. setting errors="replace"
            return tuple(
                map(
                    list,
                    zip(*{hashlib.sha256(str(doc).encode(errors="replace")).hexdigest(): doc for doc in docs}.items()),
                )
            )
        return [], []

    def add_documents(self, documents: list, **kwargs: Any) -> list[str]:
        """
        Embed and add documents to the vector store.

        Parameters
        ----------
        documents : list
            Documents to be embedded and added to the vector store.

        Returns
        -------
        list[str]
            List of documents IDs.
        """
        max_batch_size = kwargs.get("max_batch_size")
        if max_batch_size is None:
            try:
                max_batch_size = self._vector_store._client.get_max_batch_size()  # pylint: disable=protected-access
            except AttributeError:
                max_batch_size = 10_000

        ids, docs = self._process_documents(documents)
        if len(docs) > max_batch_size:
            batch_ids = []

            for batch_start in range(0, len(docs), max_batch_size):
                batch_ids.extend(
                    self._vector_store.add_documents(
                        docs[batch_start : batch_start + max_batch_size],
                        ids=ids[batch_start : batch_start + max_batch_size],
                        **kwargs,
                    )
                )
            return batch_ids

        return self._vector_store.add_documents(docs, ids=ids, **kwargs)

    def _get_window_documents(self, doc_id: str, seq_nums_window: list[int]) -> list[Document]:
        """
        Receives a document ID and a list of chunks' sequence_numbers,
        and searches the vector store according to the metadata.

        Parameters
        ----------
        doc_id : str
            ID of document.

        seq_nums_window : list[int]
            Sequence numbers of chunks.

        Returns
        -------
        list[Document]
            Documents from that document with these sequence_numbers.
        """
        expr = {
            "$and": [
                {self._document_name_field: {"$eq": doc_id}},
                {self._chunk_sequence_number_field: {"$gte": seq_nums_window[0]}},
                {self._chunk_sequence_number_field: {"$lte": seq_nums_window[-1]}},
            ]
        }
        res = self._vector_store.get(where=expr)  # type: ignore[arg-type]
        texts, metadatas = res["documents"], res["metadatas"]
        window_documents = [
            Document(
                page_content=text,
                metadata={
                    self._chunk_sequence_number_field: metadata[self._chunk_sequence_number_field],
                    self._document_name_field: doc_id,
                },
            )
            for text, metadata in zip(texts, metadatas)
        ]
        return window_documents

    _HYBRID_KWARGS = frozenset({"search_mode", "ranker_strategy", "ranker_k", "ranker_alpha"})

    def search(
        self,
        query: str,
        k: int = 5,
        include_scores: bool = False,
        **kwargs: Any,
    ) -> list[Document] | list[tuple[Document, float]]:
        """Searches for documents most similar to the query.

        The method is designed as a wrapper for respective LangChain VectorStores' similarity search methods.
        Therefore, additional search parameters passed in ``kwargs`` should be consistent with those methods,
        and can be found in the LangChain documentation.

        Parameters
        ----------
        query : str
            Query for which grounding documents will be searched for.

        k : int, default=5
            Number of documents to retrieve

        include_scores : bool, default=False
            Whether similarity scores of found documents should be returned.

        Returns
        -------
        list[Document] | list[tuple[Document, float]]
            Found documents with or without scores.
        """
        filtered_kwargs = {k_: v for k_, v in kwargs.items() if k_ not in self._HYBRID_KWARGS}
        if include_scores:
            result = self._vector_store.similarity_search_with_score(query, k=k, **filtered_kwargs)
        else:
            result = self._vector_store.similarity_search(query, k=k, **filtered_kwargs)

        return result

    def window_search(
        self,
        query: str,
        k: int = 5,
        include_scores: bool = False,
        window_size: int = 2,
        **kwargs: Any,
    ) -> list:
        """
        Searches for documents most similar to the query and extend a document (a chunk)
        to its adjacent chunks (if they exist) from the same origin document.

        The method is designed as a wrapper for respective LangChain VectorStores' similarity search methods.
        Therefore, additional search parameters passed in ``kwargs`` should be consistent with those methods,
        and can be found in the LangChain documentation.

        Parameters
        ----------
        query : str
            Query for which grounding documents will be searched for.

        k : int, default=5
            Number of documents to retrieve

        include_scores : bool, default=False
            Whether similarity scores of found documents should be returned.

        window_size : int, default=2
            Number of chunks from right and left side of the original chunk.

        Returns
        -------
        list
            Found documents with or without scores.
        """
        documents = self.search(query, k, include_scores, **kwargs)
        if window_size <= 0:
            return documents

        if not include_scores:
            documents = cast(list[Document], documents)
            return [self._window_extend_and_merge(document, window_size) for document in documents]

        documents_and_scores = cast(list[tuple[Document, float]], documents)
        documents = [t[0] for t in documents_and_scores]
        scores = [t[1] for t in documents_and_scores]
        extended_documents = [self._window_extend_and_merge(document, window_size) for document in documents]
        return list(zip(extended_documents, scores))

    def delete(self, ids: list[str], **kwargs: Any) -> None:
        """Delete by vector ID or other criteria. Sor more details see LangChain documentation
        https://python.langchain.com/api_reference/core/vectorstores/langchain_core.vectorstores.base.VectorStore.html#langchain_core.vectorstores.base.VectorStore
        """
        self._vector_store.delete(ids, **kwargs)

    def _window_extend_and_merge(self, document: Document, window_size: int) -> Document:
        """
        Extends a document (a chunk) to its adjacent chunks (if they exist) from the same origin document.
        Then merges the adjacent chunks into one chunk while keeping their order,
        and merges intersecting text between them (if it exists).
        This requires chunks to have "document_id" and "sequence_number" in their metadata.

        Parameters
        ----------
        document : Document
            Chunk / document to be extended to its window and merged.

        window_size : int
            Number of adjacent chunks to retrieve before and after the center, according to the sequence_number.

        Returns
        -------
        Document
            Chunk / document after extending and merging.
        """
        if "document_id" not in document.metadata:
            raise ValueError('document must have "document_id" in its metadata')
        if "sequence_number" not in document.metadata:
            raise ValueError('document must have "sequence_number" in its metadata')
        doc_id = document.metadata["document_id"]
        seq_num = document.metadata["sequence_number"]
        seq_nums_window = [seq_num + i for i in range(-window_size, window_size + 1, 1)]

        window_documents = self._get_window_documents(doc_id, seq_nums_window)

        window_documents.sort(key=lambda x: x.metadata["sequence_number"])
        return merge_window_into_a_document(window_documents)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import itertools
from typing import Any, Callable, TypeAlias

from ai4rag.search_space.src.default_search_space import (
    get_default_ai4rag_search_space_parameters,
)
from ai4rag.search_space.src.exceptions import SearchSpaceValueError
from ai4rag.search_space.src.parameter import Parameter
from ai4rag.utils.constants import AI4RAGParamNames, RetrievalConstraints

__all__ = ["SearchSpace", "AI4RAGSearchSpace"]


RuleFunction: TypeAlias = Callable[[dict[str, Any]], bool]


def _rule_chunk_size_bigger_than_chunk_overlap(combination: dict) -> bool:
    """Define whether combination passes selected criterion.

    Parameters
    ----------
    combination : dict
        Single node in the solutions space represented as a dict.

    Returns
    -------
    bool
        Whether combination passes selected criterion.
    """
    chunk_size = combination.get(AI4RAGParamNames.CHUNK_SIZE)
    chunk_overlap = combination.get(AI4RAGParamNames.CHUNK_OVERLAP)

    if chunk_size is None or chunk_overlap is None:
        raise SearchSpaceValueError("Chunk size and chunk overlap are required.")

    return chunk_size > 2 * chunk_overlap


def _rule_adjust_window_to_retrieval_method(combination: dict) -> bool:
    """Define whether combination passes selected criterion."""

    window_size = combination.get(AI4RAGParamNames.WINDOW_SIZE)
    retrieval_method = combination.get(AI4RAGParamNames.RETRIEVAL_METHOD)

    if retrieval_method is None or window_size is None:
        raise SearchSpaceValueError("window_size and retrieval_method are required.")

    if window_size == 0 and retrieval_method == "window":
        return False
    if window_size > 0 and retrieval_method == "simple":
        return False

    return True


def _rule_chunk_size_within_embedding_context_length(combination: dict) -> bool:
    """Check that estimated token count of a chunk fits the embedding model's context length.

    The chunk_size is in characters.  We convert to an estimated token count
    using a conservative ratio of 4.5 characters per token (i.e. we
    *overestimate* the number of tokens so that borderline cases are pruned
    rather than silently failing at runtime).

    Note: ``chunk_overlap`` is not included in the estimation because
    ``RecursiveCharacterTextSplitter`` already keeps chunks within the
    ``chunk_size`` budget — overlap is not added on top.

    Parameters
    ----------
    combination : dict
        Single node in the solutions space represented as a dict.

    Returns
    -------
    bool
        Whether the estimated token count fits within the embedding model's context length.
    """
    chunk_size = combination.get(AI4RAGParamNames.CHUNK_SIZE)
    embedding_model = combination.get(AI4RAGParamNames.EMBEDDING_MODEL)

    if chunk_size is None or embedding_model is None:
        return True

    context_length = getattr(getattr(embedding_model, "params", None), "context_length", None)
    if context_length is None:
        params = getattr(embedding_model, "params", None)
        if isinstance(params, dict):
            context_length = params.get("context_length")

    if context_length is None:
        return True

    chars_per_token = 3.6
    estimated_tokens = chunk_size / chars_per_token

    return estimated_tokens <= context_length


def _rule_search_mode_ranker_consistency(combination: dict) -> bool:
    """Ranker parameters must only be set when search_mode is 'hybrid'.

    When search_mode is 'vector', all ranker params must be sentinels
    (empty string for strategy, 0 for ranker_k, 1 for ranker_alpha).
    When search_mode is 'hybrid', ranker_strategy must be a non-empty string.

    Parameters
    ----------
    combination : dict
        Single node in the solutions space represented as a dict.

    Returns
    -------
    bool
        Whether combination passes selected criterion.
    """
    search_mode = combination.get(AI4RAGParamNames.SEARCH_MODE)
    ranker_strategy = combination.get(AI4RAGParamNames.RANKER_STRATEGY)
    ranker_k = combination.get(AI4RAGParamNames.RANKER_K)
    ranker_alpha = combination.get(AI4RAGParamNames.RANKER_ALPHA)

    if search_mode == "vector":
        if ranker_strategy or ranker_k or ranker_alpha not in (1, None):
            return False
        return True

    if search_mode == "hybrid":
        if ranker_strategy not in RetrievalConstraints.RANKER_STRATEGIES:
            return False
        return True

    return True


def _rule_ranker_k_for_rrf_only(combination: dict) -> bool:
    """ranker_k is only applicable when ranker_strategy is 'rrf'.

    For non-rrf strategies, ranker_k must be 0 (sentinel meaning unused).
    For rrf strategy, ranker_k must not be 0.

    Parameters
    ----------
    combination : dict
        Single node in the solutions space represented as a dict.

    Returns
    -------
    bool
        Whether combination passes selected criterion.
    """
    ranker_strategy = combination.get(AI4RAGParamNames.RANKER_STRATEGY)
    ranker_k = combination.get(AI4RAGParamNames.RANKER_K)

    if ranker_strategy is None or ranker_k is None:
        return True

    if ranker_strategy != "rrf" and ranker_k != 0:
        return False

    if ranker_strategy == "rrf" and ranker_k == 0:
        return False

    return True


def _rule_ranker_alpha_for_weighted_only(combination: dict) -> bool:
    """ranker_alpha is only applicable when ranker_strategy is 'weighted'.

    For non-weighted strategies, ranker_alpha must be 1 (sentinel meaning vector-only).
    For weighted strategy, ranker_alpha must not be 1.

    Parameters
    ----------
    combination : dict
        Single node in the solutions space represented as a dict.

    Returns
    -------
    bool
        Whether combination passes selected criterion.
    """
    ranker_strategy = combination.get(AI4RAGParamNames.RANKER_STRATEGY)
    ranker_alpha = combination.get(AI4RAGParamNames.RANKER_ALPHA)

    if ranker_strategy is None or ranker_alpha is None:
        return True

    if ranker_strategy != "weighted" and ranker_alpha != 1:
        return False

    if ranker_strategy == "weighted" and ranker_alpha == 1:
        return False

    return True


class SearchSpace:
    """
    Class that represents a search space used hyperparameter optimization.

    Parameters
    ----------
    params : list[Parameter]
        List of Parameters, each of which is a parameter to optimize in hyperparameter optimization process.
    """

    def __init__(self, params: list[Parameter] = None, rules: list[RuleFunction] | None = None):
        self.params = params or []
        self._search_space = {param.name: param for param in self._params}
        self._rules = rules

    def __getitem__(self, item: str) -> Parameter:
        return self._search_space[item]

    def __setitem__(self, key: str, item: Parameter) -> None:
        for idx, param in enumerate(self.params):
            if item.name == param.name:
                self.params[idx] = item
        self._search_space[item.name] = item

    def as_list(self) -> list[Parameter]:
        """
        Get the list of parameter composing the search space.

        Returns
        -------
        list[Parameter]
            List of parameters composing the search space.
        """
        return list(self.params)

    def as_dict(self) -> dict[str, Any]:
        """Return dict representation of the search space.

        Returns
        -------
        dict[str, Any]
            Dict representation of the search space."""
        return {param.name: param.all_values() for param in self._params}

    @property
    def params(self) -> list[Parameter]:
        """Get params."""
        return self._params

    @params.setter
    def params(self, params: list[Parameter]) -> None:
        """Set params."""
        if len(params) != len({param.name for param in params}):
            raise SearchSpaceValueError("Parameters must have unique names.")

        self._params = params

    @staticmethod
    def _apply_rules(combinations: list[dict], rules: list[RuleFunction]) -> list[dict]:
        """
        Apply set of rules on the given combinations.
        Remove all solutions (nodes in the space) that do not meet criteria defined in rules.

        Parameters
        ----------
        combinations : list[dict]
            Possible combinations of parameters (nodes in the space of solutions).

        rules : list[RuleFunction]
            List of rules to apply on the combinations.

        Returns
        -------
        list[dict]
            Filtered combinations of parameters after applying rules.
        """
        indexes_to_remove = []

        for idx, combination in enumerate(combinations):
            for rule in rules:
                if not rule(combination):
                    indexes_to_remove.append(idx)
                    continue

        combinations = [combination for idx, combination in enumerate(combinations) if idx not in indexes_to_remove]

        return combinations

    @property
    def combinations(self) -> list[dict]:
        """Get all possible parameters combinations."""

        space_params = {param.name: param.all_values() for param in self.params}
        combinations = [dict(zip(space_params.keys(), values)) for values in itertools.product(*space_params.values())]

        if self._rules:
            combinations = self._apply_rules(combinations, self._rules)

        return combinations

    @property
    def max_combinations(self) -> int:
        """
        Calculate how many possible combinations could be evaluated based
        on the search space.

        Returns
        -------
        int
            Number of nodes in the hyperspace.
        """
        return len(self.combinations)


class AI4RAGSearchSpace(SearchSpace):
    """
    Class that represents the search space used for the RAG hyperparameters optimization.

    Parameters
    ----------
    params : list[Parameter]
        List of Parameter, each of which is a parameter to optimize in the ai4rag process.

    rules : list[RuleFunction]
        List of functions - called "rules" - that will be applied on each combination in the search space.

    vector_store_type : str, default="ls_milvus"
        Type of vector store. When "chroma", hybrid search parameters are excluded
        from the default search space since ChromaDB does not support hybrid search.
    """

    _base_rules = (
        _rule_chunk_size_bigger_than_chunk_overlap,
        _rule_adjust_window_to_retrieval_method,
        _rule_chunk_size_within_embedding_context_length,
    )

    _hybrid_rules = (
        _rule_search_mode_ranker_consistency,
        _rule_ranker_k_for_rrf_only,
        _rule_ranker_alpha_for_weighted_only,
    )

    def __init__(
        self,
        params: list[Parameter] | None = None,
        rules: list[RuleFunction] | None = None,
        vector_store_type: str = "ls_milvus",
    ):
        default_search_space_parameters = get_default_ai4rag_search_space_parameters(vector_store_type)
        params = params or []
        self._validate_user_params(params)

        params = self._overwrite_default_search_space_with_user_provided_parameters(
            params, default_search_space_parameters
        )

        builtin_rules = self._base_rules + self._hybrid_rules if vector_store_type != "chroma" else self._base_rules
        _summed_rules = builtin_rules + rules if rules else builtin_rules
        super().__init__(params, _summed_rules)

    @staticmethod
    def _validate_user_params(params: list[Parameter]) -> None:
        """Validate parameters provided by the user, that will be later
        used for overriding the defaults.

        Parameters
        ----------
        params : list[Parameter]
            Parameters provided by the user.

        Raises
        ------
        SearchSpaceValueError
            Raised when some parameters are not recognized or required ones are missing.
        """

        required_params = (AI4RAGParamNames.FOUNDATION_MODEL, AI4RAGParamNames.EMBEDDING_MODEL)
        user_params = [param.name for param in params]
        missing_params = set(required_params) - set(user_params)

        if missing_params:
            raise SearchSpaceValueError(f"Missing required parameters in the search space: {missing_params}.")

        not_supported_params = [param for param in user_params if param not in AI4RAGParamNames]

        if not_supported_params:
            raise SearchSpaceValueError(
                f"Not supported parameters were given to the search space: {not_supported_params}."
            )

    @staticmethod
    def _overwrite_default_search_space_with_user_provided_parameters(
        params: list[Parameter],
        default_search_space_params: list[Parameter],
    ) -> list[Parameter]:
        """
        User-provided data has higher precedence than the defaults that's why we're overwriting the defaults here.

        Parameters
        ----------
        params : list[Parameter]
            List of parameters to build up this search space.

        default_search_space_params : list[Parameter]
            Default parameters to be considered.

        Returns
        -------
            Default search space overwritten and expanded (whenever necessary) with user-provided parameters.

        Raises
        ------
        SearchSpaceValueError
            When user provided unsupported parameter (not existing in the default search space).

        """
        user_params = {param.name: param for param in params}
        default_params = {param.name: param for param in default_search_space_params}

        selected_params_dict = default_params | user_params
        selected_params = list(selected_params_dict.values())

        return selected_params



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from typing import Sequence

from langchain_core.documents import Document

from ai4rag.rag.embedding.base_model import BaseEmbeddingModel

__all__ = ["BaseVectorStore"]


class BaseVectorStore(ABC):
    """
    Abstract class defining interface for VectorStore in the ai4rag experiment.
    Single instance defines 1 collection/index that can be used to store or retrieve data.
    """

    def __init__(
        self, embedding_model: BaseEmbeddingModel, distance_metric: str, reuse_collection_name: str | None = None
    ):
        self.embedding_model = embedding_model
        self.distance_metric = distance_metric
        self.reuse_collection_name = reuse_collection_name

    @abstractmethod
    def search(self, query: str, k: int, **kwargs) -> list[dict]:
        """
        Search for the chunks relevant to the query.
        The method used will be simple similarity search.

        Parameters
        ----------
        query : str
            Question / query for which the similarity search will be executed.

        k : int
            Number of chunks to be returned as a result of similarity search

        Returns
        -------
        list[dict]
            List of chunks as dicts with content and metadata.
        """

    @abstractmethod
    def add_documents(self, documents: Sequence[Document]) -> None:
        """
        Add documents to the collection.

        Parameters
        ----------
        documents : Sequence[Document]
            Documents to add to the collection.
        """

    @property
    @abstractmethod
    def collection_name(self) -> str:
        """Returns dynamically selected collection name, reused or new one."""



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from collections import deque
from collections.abc import Hashable
from dataclasses import dataclass, fields
from math import floor
from typing import Any, Generic, Literal, Optional, Protocol, Sequence, TypeVar


def _get_hashable_repr(dct: dict):
    """
    Returns
    -------
    A hashable representation of the provided dictionary.
    """
    queue = deque((k, v, 0, None) for k, v in dct.items())
    dict_unpacked = []
    while queue:
        key, val, lvl, p_ref = queue.pop()
        if hasattr(val, "items"):  # we have a nested dict
            dict_unpacked.append((key, "+", lvl, p_ref))  # key is an aggregator at this level (that's why '+')
            if hash(key) != p_ref:  # but it could be an aggregator for a Sequence (and not other dict)
                lvl += 1
            queue.extendleft((k, v, lvl, hash(key)) for k, v in val.items())
        elif isinstance(val, Hashable):
            dict_unpacked.append((key, val, lvl, p_ref))
        elif isinstance(val, Sequence):
            # only sequences supported now
            dict_unpacked.append((key, "+", lvl, p_ref))

            queue.extendleft((key, vv, floor(lvl) + ind * 0.01, hash(key)) for ind, vv in enumerate(val, 1))

        else:
            raise ValueError(f"Some value in the provided dict is not supported. {type(val)} is not supported")

    return tuple(sorted(dict_unpacked, key=lambda it: (it[2], it[0])))


class HashableProtocol(Protocol):
    """Protocol definition for hashable types."""

    def __hash__(self) -> int: ...


HashableT = TypeVar("HashableT", bound=HashableProtocol)


class ParameterValueError(ValueError):
    """Error representing incorrect value given for parameter type."""


@dataclass(frozen=True)
class Parameter(Generic[HashableT]):
    """
    Representation of the general parameter used in optimization process.
    """

    name: str
    param_type: Literal["B", "I", "R", "C"] = "C"
    v_min: Optional[int | float] = None
    v_max: Optional[int | float] = None
    values: Optional[Sequence[HashableT]] = None

    def __post_init__(self):
        """Perform basic validations for Parameter initializations."""
        param_types = ["B", "I", "R", "C"]
        if self.param_type not in param_types:
            raise ParameterValueError(f"param_type must be one of: {param_types}")

        if self.param_type in ["B", "C"]:
            if not self.values:
                raise ParameterValueError("For Boolean or Categorical types, 'values' must be provided.")
            if self.v_min is not None or self.v_max is not None:
                raise ParameterValueError(
                    "For Boolean or Categorical types, minimum and maximum values are not allowed."
                )

        if self.param_type in ["I", "R"]:
            if self.v_min is None or self.v_max is None:
                raise ParameterValueError("For Integer or Real types, minimum and maximum values must be provided.")
            if self.values is not None:
                raise ParameterValueError("For Integer or Real types, values is not supported.")
            if self.param_type == "I":
                if not (isinstance(self.v_min, int) and isinstance(self.v_max, int)):
                    raise ParameterValueError(
                        "Parameter of type categorical ('C') has to specify both 'v_max' and 'v_min' "
                        "fields as int types."
                    )

    def _hashable_field_repr_categorical_values(self) -> tuple[str | int, ...] | None:
        """
        Parameter with `param_type="C"`allow specifying a list of (possible nested) objects as available values.
        In order to reliably compare such Parameter instances between each other and ensure hashability, this
        utility function exists.
        It may seem to be computationally heavy, but for typical input its execution time still lies in
        the order of magnitude of 100ths of a milisecond.

        Returns
        -------
        tuple
            For values being a list of strings:
                sorted (string-wise, alphabetically) tuple of members in `self.values` field.
            For values being a list of dictionaries:
                a list of sorted hashes calculated for the list members
        """
        if not self.values:
            return None
        if isinstance(self.values[0], (str, bool, int, float)):
            return tuple(sorted(self.values))
        if isinstance(self.values[0], dict):
            return tuple(sorted((hash(dct_hash_repr) for dct_hash_repr in map(_get_hashable_repr, self.values))))
        # tmp workaround for param.name='inference_model_id'
        try:
            return tuple(sorted(map(hash, self.values)))
        except TypeError:
            raise ParameterValueError(
                f"Trying to get a hashable representation of values field containing "
                f"unsupported types: {type(self.values[0])}."
            ) from None

    def all_values(self) -> Sequence:
        """Get all possible values that parameter can take. It is not applicable for R."""
        match self.param_type:
            case "C":
                return self.values
            case "B":
                return [True, False]
            case "I":
                return list(range(self.v_min, self.v_max + 1))
            case _:
                raise ParameterValueError("Cannot specify possible values for R type.")

    def __eq__(self, other: Any) -> bool:
        """Parameter objects are considered equal when their fields are the same (name- and value-wise)."""
        if not isinstance(other, Parameter):
            return False

        if self._hashable_field_repr_categorical_values() != other._hashable_field_repr_categorical_values():
            return False

        for f in fields(self):
            if f.name == "values":  # already compared above
                continue
            if getattr(self, f.name) != getattr(other, f.name):
                return False

        return True

    def __hash__(self) -> int:
        if self.values:
            fields_no_values = [getattr(self, f.name) for f in fields(self) if f.name != "values"]
            return hash((*fields_no_values, self._hashable_field_repr_categorical_values()))
        return hash(tuple(getattr(self, f.name) for f in fields(self)))



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import time
from dataclasses import asdict, is_dataclass
from typing import Any, Literal, Sequence

import pandas as pd
from langchain_core.documents import Document
from llama_stack_client import LlamaStackClient

from ai4rag import logger
from ai4rag.core.experiment.benchmark_data import BenchmarkData
from ai4rag.core.experiment.exception_handler import (
    AI4RAGError,
    AssetSaveError,
    ExperimentExceptionHandler,
    IndexingError,
)
from ai4rag.core.experiment.mps import ModelsPreSelector
from ai4rag.core.experiment.results import EvaluationResult, ExperimentResults
from ai4rag.core.experiment.utils import (
    RAGExperimentError,
    RAGParamsType,
    build_evaluation_data,
    get_chunking_params,
    get_retrieval_params,
    query_rag,
)
from ai4rag.core.hpo.base_optimizer import BaseOptimizer, OptimizationError, OptimizerSettings
from ai4rag.core.hpo.gam_opt import GAMOptimizer
from ai4rag.core.hpo.random_opt import FailedIterationError
from ai4rag.evaluator.base_evaluator import BaseEvaluator, EvaluationData, MetricType
from ai4rag.evaluator.unitxt_evaluator import UnitxtEvaluator
from ai4rag.rag.chunking import LangChainChunker
from ai4rag.rag.embedding.base_model import BaseEmbeddingModel
from ai4rag.rag.foundation_models.base_model import BaseFoundationModel
from ai4rag.rag.retrieval.retriever import Retriever
from ai4rag.rag.template.simple_rag_template import SimpleRAG
from ai4rag.rag.vector_store.get_vector_store import get_vector_store
from ai4rag.search_space.src.models import EmbeddingModels
from ai4rag.search_space.src.parameter import Parameter
from ai4rag.search_space.src.search_space import AI4RAGSearchSpace
from ai4rag.utils.constants import AI4RAGParamNames, ExperimentStep
from ai4rag.utils.event_handler.event_handler import BaseEventHandler, LogLevel


# pylint: disable=too-many-instance-attributes
class AI4RAGExperiment:
    """
    Class responsible for conducting AutoRAG experiment, that consists of
    finding the best hyperparameters for several steps/stages.

    AI4RAGExperiment is essentially an orchestrator for the RAG Patterns
    hyperparameters optimization for the desired metric. It requires from
    user to provide fully defined search space on which the experiment will
    be executed.

    AI4RAG uses 'BaseRAGTemplate' inheriting classes as definitions on how
    to build and utilize RAG Pattern with the given search space nodes.

    Parameters
    ----------
    documents : list[Document | tuple[str, str]]
        List of documents to embed in vector db and use as context in RAG.
        When given as list of langchain's Document instances, both content and document
        ids must be provided:
        Document(page_content=..., metadata={document_id: 'some_id'})
        When given as list of tuples it should be (content, document_id)

    benchmark_data : pd.DataFrame | BenchmarkData
        Structure with 3 columns: 'question', 'correct_answers' and - if applicable - 'correct_answer_document_ids'.

    search_space : AI4RAGSearchSpace
        Grid of parameters used during hyperparameter optimization.

    vector_store_type : Literal["chroma", "ls_milvus"]
        Specific type of Vector Data Base that will be used during the experiment.

    optimizer_settings : OptimizerSettings
        Settings for the optimizer to be used during the experiment.

    client : LlamaStackClient | Any
        Instance of the llama stack client or other client allowing to communicate
        with the available vector store providers.

    event_handler : BaseEventHandler
        Instance satisfying BaseEventHandler's interface to stream pattern evaluation
        results and intermediate status updates. EventHandler is an entrypoint to configure
        custom logging and assets handling.

    optimization_metric : str, default=MetricType.FAITHFULNESS
        Metrics that should be used for calculating final score value that will be minimized.
        This sequence should contain 1 value for first release.


    Other Parameters
    ----------------
    job_id : str, default="ai4rag_job_a0b1c2d3"
        Unique identifier for a job.

    metrics : Sequence[str]
        Metrics that will be evaluated during AutoRAG experiment. Not all
        of these metrics will be used to calculate final score, but they will
        be included in the evaluation results.

    evaluator : BaseEvaluator, default=UnitxtEvaluator()
        An implementation of the BaseEvaluator class, that will be used by the AI4RAGExperiment
        To evaluate the RAG pattern performance and will be utilized during the optimization
        process.

    n_mps_foundation_models : int, default=3
        Amount of foundation models to be further used in experiment post pre-selection.

    n_mps_embedding_models : int, default=2
        Amount of embedding models to be further used in experiment post pre-selection.

    Attributes
    ----------
    results : ExperimentResults
        Instance holding information about each iteration during the experiment.
        It consists of statuses, RAG pattern objects, scores and settings.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        documents: list[Document],
        benchmark_data: pd.DataFrame,
        search_space: AI4RAGSearchSpace,
        vector_store_type: Literal["chroma", "ls_milvus"],
        optimizer_settings: OptimizerSettings,
        event_handler: BaseEventHandler,
        client: LlamaStackClient | Any = None,
        optimization_metric: str = MetricType.FAITHFULNESS,
        **kwargs,
    ):
        self.documents = documents
        self.benchmark_data = BenchmarkData(benchmark_data)
        self.search_space = search_space
        self.vector_store_type = vector_store_type
        self.optimizer_settings = optimizer_settings
        self.event_handler = event_handler
        self.client = client
        self.optimization_metric = optimization_metric

        self.job_id = kwargs.pop("job_id", "ai4rag_job_a0b1c2d3").replace("-", "_")
        self.metrics: Sequence[str] = kwargs.pop(
            "metrics", (MetricType.ANSWER_CORRECTNESS, MetricType.FAITHFULNESS, MetricType.CONTEXT_CORRECTNESS)
        )
        self.evaluator: BaseEvaluator = kwargs.pop(
            "evaluator",
            UnitxtEvaluator(),
        )
        self.n_mps_foundation_models = kwargs.pop("n_mps_fm", ModelsPreSelector.DEFAULT_N_FOUNDATION_MODELS)
        self.n_mps_embedding_models = kwargs.pop("n_mps_em", ModelsPreSelector.DEFAULT_N_EMBEDDING_MODELS)
        self.known_observations: list[dict] | None = kwargs.pop("known_observations", None)

        self.results: ExperimentResults = ExperimentResults()
        self._exception_handler = ExperimentExceptionHandler(self.event_handler)

        if kwargs:
            logger.warning("Unknown parameters: %s", kwargs)

    @property
    def documents(self) -> list[Document]:
        """Get list of documents"""
        return self._documents

    @documents.setter
    def documents(self, docs: list[Document | tuple[str, str]] | None) -> None:
        """
        Validate and set documents value.
        We need to make sure if we have needed content and document_ids
        provided and the documents. All documents need to be instances
        of LangChain's Document class.
        """
        proper_docs = []
        if docs:
            for idx, doc in enumerate(docs):
                if isinstance(doc, Document):
                    if not doc.page_content:
                        which_doc = doc.metadata.get("document_id") or idx
                        logger.warning("Empty document: %s", which_doc)
                    if not doc.metadata.get("document_id", None):
                        logger.warning("document_id not provided for document at index: %s", idx)

                    proper_docs.append(doc)

                else:
                    raise ValueError(f"Incorrect type of document provided at index: {idx}.")

        self._documents = proper_docs

    @property
    def optimization_metric(self) -> str:
        """Get optimization metrics used for the experiment."""
        return self._optimization_metric

    @optimization_metric.setter
    def optimization_metric(self, val: str) -> None:
        """Validate and set optimization metrics"""
        if val not in MetricType:
            raise RAGExperimentError(
                f"Provided optimization metric: '{val}' is not supported. "
                f"Available metrics: ['answer_correctness', 'faithfulness', 'context_correctness']."
            )

        self._optimization_metric = val

    @property
    def benchmark_data(self) -> BenchmarkData:
        """Get benchmark data."""
        return self._benchmark_data

    @benchmark_data.setter
    def benchmark_data(self, val: BenchmarkData) -> None:
        """Check and set benchmark data based on the executed scenario."""
        self._benchmark_data = val

    def run_pre_selection(
        self,
        foundation_models: list[BaseFoundationModel],
        embedding_models: list[BaseEmbeddingModel],
        n_records: int = 5,
        random_seed: int = 17,
    ) -> dict[str, list[BaseEmbeddingModel | BaseFoundationModel]]:
        """
        Run models pre-selection using ModelsPreSelector and sample
        of the data.

        Parameters
        ----------
        embedding_models : list[BaseEmbeddingModel]
            Embedding models to be considered during pre-selection process.

        foundation_models : list[BaseFoundationModel]
            Foundation models to be evaluated during pre-selection process.

        n_records : int, default=5
            Amount of records that should be used during models pre-selection.

        random_seed : int, default=17
            Random seed value used for sampling benchmark data records.

        Returns
        -------
        dict[str, list[BaseFoundationModel | EmbeddingModel]]
            Best embedding models and foundation models found in pre-selection.
        """
        _log_start_mps = (
            f"Starting foundation models pre-selection with following "
            f"foundation models: {[str(fm) for fm in foundation_models]} "
            f"and following embedding models: {[str(em) for em in embedding_models]}."
        )
        logger.info(_log_start_mps)
        self.event_handler.on_status_change(
            level=LogLevel.INFO,
            message=_log_start_mps,
            step=ExperimentStep.MODEL_SELECTION,
        )

        mps = ModelsPreSelector(
            benchmark_data=self.benchmark_data.get_random_sample(n_records=n_records, random_seed=random_seed),
            documents=self.documents.copy(),
            foundation_models=foundation_models,
            embedding_models=embedding_models,
            metric=self.optimization_metric,
        )
        mps.evaluate_patterns()

        selected_models = mps.select_models(
            n_embedding_models=self.n_mps_embedding_models, n_foundation_models=self.n_mps_foundation_models
        )

        logger.info(
            "Models pre-selection has been finished. Selected foundation models: %s and selected embedding models: %s.",
            [str(model) for model in selected_models["foundation_models"]],
            [str(model) for model in selected_models["embedding_models"]],
        )

        return selected_models

    # pylint: disable=too-many-locals, too-many-statements
    def run_single_evaluation(self, rag_params: RAGParamsType) -> float:
        """
        Evaluate a single RAG configuration and return its score using provided documents.

        Parameters
        ----------
        rag_params : RAGParamsType
            A dictionary containing rag parameters as keys and their values.

        Returns
        -------
        float
            A single evaluation score obtained by the executed rag pattern.
        """
        start_time = time.time()

        chunking_params = get_chunking_params(rag_params)
        retrieval_params = get_retrieval_params(rag_params)

        foundation_model = rag_params.get(AI4RAGParamNames.FOUNDATION_MODEL)
        embedding_model = rag_params.get(AI4RAGParamNames.EMBEDDING_MODEL)

        distance_metric = EmbeddingModels.get_distance_metric(embedding_model.model_id)
        embedding_params_dict = (
            asdict(embedding_model.params) if is_dataclass(embedding_model.params) else embedding_model.params
        )
        indexing_params = {
            "chunking": chunking_params,
            "embedding": {
                "model_id": embedding_model.model_id,
                "distance_metric": distance_metric,
                "embedding_params": embedding_params_dict,
            },
        }

        logger.info("Using indexing params: %s", indexing_params)

        retrieval_method = retrieval_params[AI4RAGParamNames.RETRIEVAL_METHOD]
        number_of_chunks = retrieval_params[AI4RAGParamNames.NUMBER_OF_CHUNKS]

        search_mode = retrieval_params.get(AI4RAGParamNames.SEARCH_MODE, "vector")
        if search_mode != "vector" and self.vector_store_type == "chroma":
            raise RAGExperimentError(
                f"Search mode '{search_mode}' is not supported with chroma vector store. "
                "Only 'vector' mode is supported for chroma."
            )

        context_template_text = foundation_model.context_template_text
        system_message_text = foundation_model.system_message_text
        user_message_text = foundation_model.user_message_text

        rag_params = {
            "retrieval": retrieval_params,
            "generation": {
                "model_id": foundation_model.model_id,
                "context_template_text": context_template_text,
                "user_message_text": user_message_text,
                "system_message_text": system_message_text,
            },
        }

        logger.info("Using retrieval and generation params: %s", rag_params)

        result_score = self.results.evaluation_explored_or_cached(
            indexing_params=indexing_params, rag_params=rag_params
        )
        if result_score is not None:
            return result_score

        pattern_name = self._create_pattern_name()
        logger.info("Using name '%s' for the currently evaluated pattern.", pattern_name)

        reuse_collection_name = self._get_reusable_collection_name(indexing_params=indexing_params)

        vector_store = get_vector_store(
            vs_type=self.vector_store_type,
            embedding_model=embedding_model,
            reuse_collection_name=reuse_collection_name,
            client=self.client,
        )

        collection_name = vector_store.collection_name

        if not self._collection_exists(collection_name=collection_name):
            chunking_method = chunking_params.get(AI4RAGParamNames.CHUNKING_METHOD)
            chunk_size = chunking_params.get(AI4RAGParamNames.CHUNK_SIZE)
            chunk_overlap = chunking_params.get(AI4RAGParamNames.CHUNK_OVERLAP)

            chunker = LangChainChunker(method=chunking_method, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
            chunked_documents = chunker.split_documents(self.documents)

            if self.event_handler:
                self.event_handler.on_status_change(
                    level=LogLevel.INFO,
                    message=(
                        f"Chunking documents using the {chunking_method} method, chunk_size: {chunk_size} "
                        f"and chunk_overlap: {chunk_overlap}."
                    ),
                    step=ExperimentStep.CHUNKING,
                )

            self.event_handler.on_status_change(
                level=LogLevel.INFO,
                message=(
                    f"Embedding chunks using the {embedding_model.model_id} model. "
                    f"Building index: {collection_name}."
                ),
                step=ExperimentStep.EMBEDDING,
            )

            try:
                vector_store.add_documents(chunked_documents)
            except Exception as exc:
                raise IndexingError(exc, collection_name, embedding_model.model_id) from exc

        else:
            self.event_handler.on_status_change(
                level=LogLevel.INFO,
                message=f"Using index {collection_name}.",
                step=ExperimentStep.EMBEDDING,
            )

        logger.info("Using retriever with parameters: %s", retrieval_params)

        retriever = Retriever(
            vector_store=vector_store,
            number_of_chunks=number_of_chunks,
            method=retrieval_method,
            search_mode=search_mode,
            ranker_strategy=retrieval_params.get(AI4RAGParamNames.RANKER_STRATEGY),
            ranker_k=retrieval_params.get(AI4RAGParamNames.RANKER_K),
            ranker_alpha=retrieval_params.get(AI4RAGParamNames.RANKER_ALPHA),
        )

        rag_pattern = SimpleRAG(
            foundation_model=foundation_model,
            retriever=retriever,
        )

        _rag_log = (
            f"Retrieval and generation using collection: '{collection_name}' and "
            f"foundation model: '{foundation_model.model_id}'."
        )
        logger.info(_rag_log)
        self.event_handler.on_status_change(
            level=LogLevel.INFO,
            message=_rag_log,
            step=ExperimentStep.GENERATION,
        )

        inference_response = query_rag(
            rag=rag_pattern,
            questions=list(self.benchmark_data.questions),
        )
        result_scores, evaluation_data = self._evaluate_response(
            inference_response=inference_response,
            pattern_name=pattern_name,
        )

        stop_time = time.time()
        execution_time = stop_time - start_time

        result_score = result_scores["scores"][self.optimization_metric]["mean"]

        logger.info("Calculated optimization score for '%s': %s", pattern_name, result_score)

        evaluation_result = EvaluationResult(
            pattern_name=pattern_name,
            collection=collection_name,
            indexing_params=indexing_params,
            rag_params=rag_params,
            scores=result_scores,
            execution_time=execution_time,
            final_score=result_score,
            rag_pattern=rag_pattern,
        )

        evaluation_results_json = self.results.create_evaluation_results_json(
            evaluation_data=evaluation_data, evaluation_result=evaluation_result
        )

        logger.info(
            "Evaluation scores: %s",
            {el.get("question_id"): el.get("scores") for el in evaluation_results_json if isinstance(el, dict)},
        )

        try:
            self._stream_finished_pattern(
                evaluation_result=evaluation_result,
                evaluation_results_json=evaluation_results_json,
            )
        except Exception as exc:
            raise AssetSaveError(exc) from exc

        self.results.add_evaluation(
            evaluation_data=evaluation_data,
            evaluation_result=evaluation_result,
        )

        return result_score

    def search(self, **kwargs) -> None:
        """
        Prepare and execute experiment to find the best RAG parameters.

        Result of the search() can be reviewed via self.results as this object
        stores results of each evaluation or via self.event_handler with custom
        implementation.
        """

        logger.info("Starting RAG optimization process...")

        def objective_function(space: RAGParamsType) -> float | None:
            """Function passed to the optimizer."""
            try:
                return self.run_single_evaluation(space)
            except AI4RAGError as err:
                msg = self._exception_handler.handle_exception(err)
                raise FailedIterationError(msg) from err

        # MPS - models pre-selection based on sample evaluation.
        # Run if there are more than 3 foundation models or more than 2 embedding models.
        foundation_models = list(self.search_space[AI4RAGParamNames.FOUNDATION_MODEL].values)
        embedding_models = list(self.search_space[AI4RAGParamNames.EMBEDDING_MODEL].values)

        if (
            len(embedding_models) > self.n_mps_embedding_models or len(foundation_models) > self.n_mps_foundation_models
        ) and not kwargs.get("skip_mps", False):
            selected_models = self.run_pre_selection(
                foundation_models=foundation_models, embedding_models=embedding_models
            )
            self.search_space[AI4RAGParamNames.FOUNDATION_MODEL] = Parameter(
                name=AI4RAGParamNames.FOUNDATION_MODEL, param_type="C", values=selected_models["foundation_models"]
            )
            self.search_space[AI4RAGParamNames.EMBEDDING_MODEL] = Parameter(
                name=AI4RAGParamNames.EMBEDDING_MODEL, param_type="C", values=selected_models["embedding_models"]
            )

        optimizer_class: type[BaseOptimizer] = kwargs.get("optimizer", GAMOptimizer)

        optimizer_kwargs = {}
        if self.known_observations is not None:
            optimizer_kwargs["known_observations"] = self.known_observations

        # In the search kwargs user may pass different optimizer class for testing purposes
        optimizer = optimizer_class(
            objective_function=objective_function,
            search_space=self.search_space,
            settings=self.optimizer_settings,
            **optimizer_kwargs,
        )
        logger.info(
            "Using optimizer: %s with optimizer settings: %s",
            optimizer_class.__name__,
            self.optimizer_settings.to_dict(),
        )

        try:
            _ = optimizer.search()
        except OptimizationError as err:
            final_error_msg = self._exception_handler.get_final_error_msg()
            raise RAGExperimentError(final_error_msg) from err

        self.event_handler.on_status_change(
            level=LogLevel.INFO,
            message="Experiment optimization process finished.",
        )

    def _stream_finished_pattern(
        self,
        evaluation_result: EvaluationResult,
        evaluation_results_json: list,
    ) -> None:
        """
        Stream finished pattern.

        Parameters
        ----------
        evaluation_result : EvaluationResult
            Data made of evaluation results.

        evaluation_results_json : list
            Prepared partial payload for the streamed content.
        """
        retrieval_payload = {
            "method": evaluation_result.rag_params["retrieval"][AI4RAGParamNames.RETRIEVAL_METHOD],
            "number_of_chunks": evaluation_result.rag_params["retrieval"][AI4RAGParamNames.NUMBER_OF_CHUNKS],
            "search_mode": evaluation_result.rag_params["retrieval"].get(AI4RAGParamNames.SEARCH_MODE, "vector"),
        }

        if evaluation_result.rag_params["retrieval"][AI4RAGParamNames.WINDOW_SIZE]:
            retrieval_payload["window_size"] = evaluation_result.rag_params["retrieval"][AI4RAGParamNames.WINDOW_SIZE]

        if retrieval_payload["search_mode"] == "hybrid":
            retrieval_payload["ranker_strategy"] = evaluation_result.rag_params["retrieval"].get(
                AI4RAGParamNames.RANKER_STRATEGY
            )
            retrieval_payload["ranker_k"] = evaluation_result.rag_params["retrieval"].get(AI4RAGParamNames.RANKER_K)
            retrieval_payload["ranker_alpha"] = evaluation_result.rag_params["retrieval"].get(
                AI4RAGParamNames.RANKER_ALPHA
            )

        vector_store_payload = {
            "datasource_type": self.vector_store_type,
            "collection_name": evaluation_result.collection,
        }

        indexing_payload = {
            "chunking": {
                "method": evaluation_result.indexing_params["chunking"][AI4RAGParamNames.CHUNKING_METHOD],
                "chunk_size": evaluation_result.indexing_params["chunking"][AI4RAGParamNames.CHUNK_SIZE],
                "chunk_overlap": evaluation_result.indexing_params["chunking"][AI4RAGParamNames.CHUNK_OVERLAP],
            },
            "embedding": evaluation_result.indexing_params.get("embedding"),
        }

        generation_payload = evaluation_result.rag_params.get("generation")

        n_known = len(self.known_observations) if self.known_observations else 0

        payload = {
            "pattern_name": evaluation_result.pattern_name,
            "scores": {
                "scores": evaluation_result.scores["scores"],
                "question_scores": evaluation_result.scores["question_scores"],
            },
            "execution_time": int(evaluation_result.execution_time),
            "final_score": evaluation_result.final_score,
            "schema_version": "1.0",
            "producer": "ai4rag",
            "settings": {
                "vector_store": vector_store_payload,
                **indexing_payload,
                "retrieval": retrieval_payload,
                "generation": generation_payload,
            },
            "iteration": len(self.results) + n_known,
        }

        self.event_handler.on_pattern_creation(
            payload=payload,
            evaluation_results=evaluation_results_json,
        )

    def _evaluate_response(
        self,
        inference_response: list[dict[str, Any]],
        pattern_name: str,
    ) -> tuple[dict[str, dict], list[EvaluationData]]:
        """
        Evaluate response from the model based on the chosen context,
        real questions/answers/ids from the benchmark_data.

        Parameters
        ----------
        inference_response : list[dict[str, Any]]
            List of model's responses containing question, answer and used
            reference documents for each record.

        pattern_name : str
            Name of the pattern for which evaluation is performed.

        Returns
        -------
        tuple[dict[str, dict], list[EvaluationData]]
            Data from evaluation that is of following structure:
            data = {
                "scores": {"answer_correctness": {"mean": 0.5, "ci_low": 0.4, "ci_high": 0.6}, ...},
                "question_scores": {
                    "answer_correctness": {"q_id_0": 0.5, "q_id_1": 0.8, ...},
                    "context_correctness": {"q_id_0": 0.5, "q_id_1": 0.8, ...},
                },
            }
        """

        logger.info(
            "Evaluating the RAG Pattern '%s' response using %s.", pattern_name, self.evaluator.__class__.__name__
        )
        self.event_handler.on_status_change(
            level=LogLevel.INFO,
            message=f"Evaluating the RAG Pattern '{pattern_name}' response using {self.evaluator.__class__.__name__}.",
            step="evaluation",
        )

        eval_data = build_evaluation_data(benchmark_data=self.benchmark_data, inference_response=inference_response)
        result = self.evaluator.evaluate_metrics(evaluation_data=eval_data, metrics=self.metrics)

        logger.info("Response evaluation results for '%s': %s.", pattern_name, result)
        return result, eval_data

    def _collection_exists(self, collection_name: str) -> bool:
        """
        This method checks if a collection with a given name already exists.
        The trick comes with chromadb. We always need to assume that collection
        does not exist, as we create new instance of chroma in memory per each
        run.

        Parameters
        ----------
        collection_name : str
            Name of the collection to check if exists.

        Returns
        -------
        bool
            True if collection exist, otherwise False.
        """
        return collection_name in self.results.collection_names

    def _get_reusable_collection_name(self, indexing_params: dict[str, Any]) -> str | None:
        """
        This method returns name of the collection / vector_store_id (for llama stack)
        if chosen indexing params have already been used to create an index / collection.

        Parameters
        ----------
        indexing_params : dict[str, Any]
            Dictionary containing keys and values that are compared with
            previously used ones, to establish if the newly created collection
            would be exactly the same.

        Returns
        -------
        str | None
            Collection name that is new or one of the previously created.
            None if there is no collection to reuse.
        """
        collection = self.results.get_existing_collection(indexing_params=indexing_params)
        if collection is not None:
            collection_name = collection
            logger.info("Reusing existing collection: '%s'", collection_name)
            return collection_name

        return None

    def _create_pattern_name(self) -> str:
        """
        Create pattern name based on the already existing patterns and length
        of results, as we iterate patterns from 1 to n.

        Returns
        -------
        str
            Pattern name.
            Example: "Pattern7"
        """
        n_known = len(self.known_observations) if self.known_observations else 0
        return f"Pattern{n_known + len(self.results) + 1}"



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from typing import Generic, TypedDict, TypeVar

from ai4rag.rag.foundation_models.utils import RAGPromptTemplateString
from ai4rag.search_space.src.model_props import (
    get_context_template_text,
    get_system_message_text,
    get_user_message_text,
)

FoundationModelClientT = TypeVar("FoundationModelClientT")
FoundationModelParamsT = TypeVar("FoundationModelParamsT")


class MessageTyped(TypedDict):
    """Type of the messages used by the client."""

    role: str
    content: str


class BaseFoundationModel(Generic[FoundationModelClientT, FoundationModelParamsT], ABC):
    """Interface definition for the foundation model used for `ai4rag`."""

    user_message_text: RAGPromptTemplateString = RAGPromptTemplateString(template_name="user_message_text")
    context_template_text: RAGPromptTemplateString = RAGPromptTemplateString(template_name="context_template_text")

    def __init__(
        self,
        client: FoundationModelClientT,
        model_id: str,
        params: FoundationModelParamsT,
        system_message_text: str | None = None,
        user_message_text: str | None = None,
        context_template_text: str | None = None,
    ):
        self.client = client
        self.model_id = model_id
        self.params = params
        self.system_message_text = system_message_text or get_system_message_text(model_name=model_id)
        self.user_message_text = (
            user_message_text if user_message_text is not None else get_user_message_text(model_name=model_id)
        )
        self.context_template_text = (
            context_template_text
            if context_template_text is not None
            else get_context_template_text(model_name=model_id)
        )

    def __repr__(self) -> str:
        return self.model_id

    def __str__(self) -> str:
        return self.model_id

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseFoundationModel):
            return NotImplemented

        return self.model_id == other.model_id

    def __hash__(self):
        return hash(self.model_id)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, BaseFoundationModel):
            return NotImplemented
        return self.model_id < other.model_id

    @abstractmethod
    def chat(self, messages: list[MessageTyped]) -> list[MessageTyped]:
        """Chat with the model base on the client capabilities."""



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Callable

from ai4rag.search_space.src.search_space import SearchSpace

__all__ = ["BaseOptimizer", "OptimizerSettings", "OptimizationError", "FailedIterationError"]


class OptimizationError(Exception):
    """Custom class representing exception occurring in the Optimizer."""


class FailedIterationError(Exception):
    """Error used to signalize failed iteration in the experiment."""


@dataclass
class OptimizerSettings:
    """
    Representation of the general Optimizer Settings.

    Parameters
    ----------
    max_evals : int
        Maximum number of evaluations performed during optimization process.

    Methods
    -------
    to_dict()
        Cast all the dataclass into the dictionary
    """

    max_evals: int

    def to_dict(self) -> dict:
        """
        Cast settings to dictionary

        Returns
        -------
        dict
            Dictionary representation of the settings class.
        """
        return asdict(self)


class BaseOptimizer(ABC):
    """
    Abstract class defining interface of Optimizer used in AI4RAGExperiment

    Parameters
    ----------
    objective_function : Callable[[dict], float]
        Target function that will be used in every evaluation. Output of
        this function should be 'float', as this is the value that algorithms
        try to minimize. Function should take dict filled with 'key: value' pairs
        that are 'argument: corresponding value'.

    search_space : SearchSpace
        List of parameters that algorithm will optimize.

    settings : OptimizerSettings
        Instance holding all the settings needed for the user

    Methods
    -------
    search()
        Perform hyperparameter optimization to find what point in search space
        gives the minimal value for the objective function.
    """

    def __init__(
        self,
        objective_function: Callable[[dict], float],
        search_space: SearchSpace,
        settings: OptimizerSettings,
    ):
        self.objective_function = objective_function
        self._search_space = search_space
        self.settings = settings

    @abstractmethod
    def search(self) -> dict[str, Any]:
        """Start process of exploring parameters space to find the best combination and the target value."""



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

ClientT = TypeVar("ClientT")
EmbeddingParamsT = TypeVar("EmbeddingParamsT")


class BaseEmbeddingModel(ABC, Generic[ClientT, EmbeddingParamsT]):
    """Interface definition for Embedding Model that will be used for `ai4rag`."""

    def __init__(self, client: ClientT, model_id: str, params: EmbeddingParamsT | None = None):
        self.client: ClientT = client
        self.model_id = model_id
        self.params: EmbeddingParamsT = params

    def __str__(self) -> str:
        return self.model_id

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseEmbeddingModel):
            return NotImplemented

        return self.model_id == other.model_id

    def __hash__(self):
        return hash(self.model_id)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, BaseEmbeddingModel):
            return NotImplemented
        return self.model_id < other.model_id

    @abstractmethod
    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embed documents.

        Parameters
        ----------
        texts : list[str]
            List of text-like chunks.

        Returns
        -------
        list[list[float]]
            Embeddings made from the list of texts.
        """

    @abstractmethod
    def embed_query(self, query: str) -> list[float]:
        """Embed query text

        Parameters
        ----------
        query : str
            User's query as text.

        Returns
        -------
        list[float]
            Single embeddings vector made from the query.
        """



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------

from typing import Any

from langchain_core.documents import Document

from ai4rag.rag.chunking.langchain_chunker import LangChainChunker
from ai4rag.rag.retrieval.retriever import Retriever
from ai4rag.rag.vector_store.llama_stack import LSVectorStore

from ..embedding.base_model import BaseEmbeddingModel
from ..foundation_models.base_model import BaseFoundationModel
from .base_template import BaseRAGTemplate, RAGTemplateError


class SimpleRAG(BaseRAGTemplate):
    """
    RAG template using Llama Stack components for embedding, vector store,
    retrieval, and foundation model, with LangChain for document chunking.

    This template implements the BaseRAGTemplate using Llama Stack services
    for all major RAG components while utilizing LangChain's chunking capabilities.

    Parameters
    ----------
    foundation_model : BaseFoundationModel
        Initialized Llama Stack foundation model for text generation.

    retriever : Retriever
        Initialized retriever for document retrieval.

    chunker : LangChainChunker | None, default=None
        Initialized LangChain chunker for document splitting.

    embedding_model : LSEmbeddingModel | None, default=None
        Initialized Llama Stack embedding model.

    vector_store : LSVectorStore | None, default=None
        Initialized Llama Stack vector store.
    """

    def __init__(
        self,
        foundation_model: BaseFoundationModel,
        retriever: Retriever,
        chunker: LangChainChunker | None = None,
        embedding_model: BaseEmbeddingModel | None = None,
        vector_store: LSVectorStore | None = None,
    ):
        super().__init__(
            foundation_model=foundation_model,
            retriever=retriever,
            embedding_model=embedding_model,
            vector_store=vector_store,
        )

        self.chunker = chunker

    def build_index(self, documents: list[Document], **kwargs) -> None:
        """
        Index documents into the vector store.

        This method chunks the documents using the LangChain chunker and
        adds them to the vector store.

        Parameters
        ----------
        documents : list[Document]
            List of LangChain Document objects to index.
        """
        if self.chunker is None and self.embedding_model is None and self.vector_store is None:
            raise RAGTemplateError()
        chunks = self.chunker.split_documents(documents)

        self.vector_store.add_documents(chunks)

    def generate(self, question: str, **kwargs) -> dict[str, Any]:
        """
        Generate an answer for a question using RAG pipeline.

        Parameters
        ----------
        question : str
            The user's question.

        **kwargs
            Additional parameters (e.g., number_of_chunks).

        Returns
        -------
        dict[str, Any]
            Dictionary containing:
            - "answer": The generated answer
            - "reference_documents": The retrieved document chunks
            - "question": The original question
        """
        reference_documents = self.retriever.retrieve(question, **kwargs)

        context = "\n".join(
            [
                self.foundation_model.context_template_text.format(document=getattr(doc, "page_content", ""))
                for doc in reference_documents
            ]
        )

        user_message = self.foundation_model.user_message_text.format(
            reference_documents=context,
            question=question,
        )

        messages = [
            {"role": "system", "content": self.foundation_model.system_message_text},
            {"role": "user", "content": user_message},
        ]

        chat_response = self.foundation_model.chat(messages=messages)

        return {
            "answer": chat_response[0].message.content,
            "reference_documents": reference_documents,
            "question": question,
        }

    def generate_stream(self, question: str, **kwargs):
        """
        Generate a streaming answer for a question using RAG pipeline.

        Note: This is a placeholder implementation. Full streaming support
        would require streaming capabilities in the LSFoundationModel.

        Parameters
        ----------
        question : str
            The user's question.

        **kwargs
            Additional parameters (e.g., number_of_chunks).

        Yields
        ------
        str
            Chunks of the generated answer.
        """
        result = self.generate(question, **kwargs)
        yield result["answer"]



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any, Iterable, Literal, Sequence

from langchain_core.documents import Document
from langchain_text_splitters import RecursiveCharacterTextSplitter, TextSplitter

from .base_chunker import BaseChunker

__all__ = [
    "LangChainChunker",
]


class LangChainChunker(BaseChunker[Document]):
    """
    Wrapper for LangChain TextSplitter.

    Parameters
    ----------
    method : Literal["recursive", "character", "token"], default="recursive"
        Describes the type of TextSplitter as the main instance performing the chunking.

    chunk_size : int, default=2048
        Maximum size of a single chunk that is returned.

    chunk_overlap : int, default=256
        Overlap in characters between chunks.

    Other Parameters
    ----------------
    separators : list[str]
        Separators between chunks.
    """

    supported_methods = ("recursive",)

    def __init__(
        self,
        method: Literal["recursive", "character", "token"] = "recursive",
        chunk_size: int = 2048,
        chunk_overlap: int = 256,
        **kwargs: Any,
    ) -> None:
        self.method = method
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = kwargs.pop("separators", ["\n\n", r"(?<=\. )", "\n", " ", ""])
        self._text_splitter = self._get_text_splitter()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LangChainChunker):
            return self.to_dict() == other.to_dict()
        return NotImplemented

    def _get_text_splitter(self) -> TextSplitter:
        """Create an instance of TextSplitter based on the settings."""

        match self.method:
            case "recursive":

                text_splitter = RecursiveCharacterTextSplitter(
                    chunk_size=self.chunk_size,
                    chunk_overlap=self.chunk_overlap,
                    separators=self.separators,
                    length_function=len,
                    add_start_index=True,
                )

            case _:
                raise ValueError(
                    f"Chunker method '{self.method}' is not supported. Use one of {self.supported_methods}."
                )

        return text_splitter

    def to_dict(self) -> dict[str, Any]:
        """
        Return dictionary that can be used to recreate an instance of the LangChainChunker.
        """
        params = (
            "method",
            "chunk_size",
            "chunk_overlap",
        )

        ret = {k: v for k, v in self.__dict__.items() if k in params}

        return ret

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "LangChainChunker":
        """Create an instance from the dictionary."""

        return cls(**d)

    @staticmethod
    def _set_document_id_in_metadata_if_missing(documents: Iterable[Document]) -> None:
        """
        Sets "document_id" in the metadata if it is missing.
        The document_id is the hash of the document's content.

        Parameters
        ----------
        documents : Iterable[Document]
            Sequence of documents for which document ids will be provided.
        """
        for doc in documents:
            if "document_id" not in doc.metadata:
                doc.metadata["document_id"] = str(hash(doc.page_content))

    @staticmethod
    def _set_sequence_number_in_metadata(chunks: list[Document]) -> list[Document]:
        """
        Sets "sequence_number" in the metadata, sorted by chunks' "start_index".

        Parameters
        ----------
        chunks : list[Document]
            Sequence of chunks of documents that contain context in a text format.

        Returns
        -------
        list[Document]
            List of updated chunks, sorted by document_id and sequence_number.
        """
        # sort chunks by start_index for each document_id
        sorted_chunks = sorted(chunks, key=lambda x: (x.metadata["document_id"], x.metadata["start_index"]))

        document_sequence: dict[str, int] = {}
        for chunk in sorted_chunks:
            doc_id = chunk.metadata["document_id"]
            prev_seq_num = document_sequence.get(doc_id, 0)
            seq_num = prev_seq_num + 1
            document_sequence[doc_id] = seq_num
            chunk.metadata["sequence_number"] = seq_num

        return sorted_chunks

    def split_documents(self, documents: Sequence[Document]) -> list[Document]:
        """
        Split series of documents into smaller chunks based on the provided
        chunker settings. Each chunk has metadata that includes the document_id,
        sequence_number, and start_index.

        Parameters
        ----------
        documents : Sequence[Document]
            Sequence of elements that contain context in a text format.

        Returns
        -------
        list[Document]
            List of documents split into smaller chunks.
        """
        self._set_document_id_in_metadata_if_missing(documents)
        chunks = self._text_splitter.split_documents(documents)
        sorted_chunks = self._set_sequence_number_in_metadata(chunks)
        return sorted_chunks



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from dataclasses import asdict, dataclass
from typing import Any, Generator

from ai4rag import logger
from ai4rag.evaluator.base_evaluator import EvaluationData
from ai4rag.rag.template.base_template import BaseRAGTemplate


@dataclass(frozen=True)
class EvaluationResult:
    """
    Class holding data from single evaluation.

    Parameters
    ----------
    pattern_name : str
        Name of the RAG pattern created. Next names should be
        i.e. Pattern1, Pattern2, ..., PatternX

    collection : list[str]
        Name of the indexes/collections in the VectorStores for single
        configuration.

    indexing_params : dict[str, Any] | None
        Subspace of hyperparameters used during indexing stage in the
        Retrieval Augmented Generation.

    rag_params : dict[str, Any]
        Subspace of hyperparameters used during inference stage in the
        Retrieval Augmented Generation.

    scores : dict[str, dict[str, float]]
        Score data from the evaluation that may look like:
        data = {
            "scores": {"answer_correctness": {"mean": 0.5, "ci_low": 0.4, "ci_high": 0.6}, ...},
            "question_scores": {
                "answer_correctness": {"q_id_0": 0.5, "q_id_1": 0.8, ...},
                "context_correctness": {"q_id_0": 0.5, "q_id_1": 0.8, ...},
            },
        }

    execution_time : float
        Time in seconds how long did experiment take to run.

    final_score : float
        Single score calculated for optimization process as the value to be minimized or maximized.

    rag_pattern : BaseRAGTemplate | None, default=None
        RAG Pattern instance that can be used.
    """

    pattern_name: str
    collection: str
    indexing_params: dict[str, Any]
    rag_params: dict[str, Any]
    scores: dict[str, dict]
    execution_time: float
    final_score: float
    rag_pattern: BaseRAGTemplate | None = None

    def to_dict(self) -> dict[str, Any]:
        """Cast instance to dict"""
        return asdict(self)


class ExperimentResults:
    """
    Class holding information about each run in the optimization process.

    Attributes
    ----------
    evaluations : list[EvaluationResult]
        Instances of each evaluation.

    evaluation_data : list[list[EvaluationData]]
        Data for evaluation, questions, ground truth answers and context references.

    Methods
    -------
    add_evaluation()
        Extend space with another evaluation step from the objective function.

    sorted_(reverse=False)
        Return sorted version of the evaluations.

    get_best_evaluations(k=None)
        Return k-best evaluations.
        If k is None all evaluations are returned sorted based on optimization metric score.
    """

    def __init__(self):
        self.evaluations: list[EvaluationResult] = []
        self.evaluation_data: list[list[EvaluationData]] = []

    def __len__(self) -> int:
        return len(self.evaluations)

    def __iter__(self) -> Generator[EvaluationResult, None, None]:
        yield from self.evaluations

    def __bool__(self) -> bool:
        return bool(self.evaluations)

    def add_evaluation(
        self,
        evaluation_data: list[EvaluationData],
        evaluation_result: EvaluationResult,
    ) -> None:
        """
        This method is responsible for adding searched point in the parameters space.

        Parameters
        ----------
        evaluation_data : list[EvaluationData]
            Data used to make evaluation. It is a list, as one instance
            of EvaluationData is per 1 question and answer.

        evaluation_result : EvaluationResult
            Instance containing information about single HPO run.
        """
        logger.info("Adding evaluation results and evaluation data to training results.")
        self.evaluations.append(evaluation_result)
        self.evaluation_data.append(evaluation_data)

    def evaluation_explored_or_cached(
        self, indexing_params: dict[str, Any], rag_params: dict[str, Any]
    ) -> float | None:
        """
        This method checks if an evaluation to certain params already exists.

        Parameters
        ----------
        indexing_params : dict[str, Any]
            Dictionary containing keys and values that are compared with
            previously used ones, to establish if the evaluation already done.

        rag_params : dict[str, Any]
            Dictionary containing keys and values that are compared with
            previously used ones, to establish if the evaluation already done.

        Returns
        -------
        float | None
            The final score of the evaluation if it exists, otherwise None.
        """
        for evaluation in self.evaluations:
            if evaluation.indexing_params == indexing_params and evaluation.rag_params == rag_params:
                return evaluation.final_score
        return None

    def get_existing_collection(self, indexing_params: dict[str, Any]) -> str | None:
        """
        This method checks if a collection to certain params already exists.
        If so, returns the collection name.

        Parameters
        ----------
        indexing_params : dict[str, Any]
            Dictionary containing keys and values that are compared with
            previously used ones, to establish if the index already created.

        Returns
        -------
        str | None
            The collection name of the index if it exists, otherwise None.
        """
        for evaluation in self.evaluations:
            if evaluation.indexing_params == indexing_params:
                return evaluation.collection
        return None

    @property
    def collection_names(self) -> list[str]:
        """
        Return list of collection_names from the HPO run.

        Returns
        -------
        list[str]
            Collections/indexes names created during all HPO runs.
        """
        ret = set()
        for ev in self.evaluations:
            ret.add(ev.collection)
        return list(ret)

    def get_best_evaluations(self, k: int | None = None) -> tuple[EvaluationResult]:
        """
        Return k-best evaluations.

        Parameters
        ----------
        k : int
            how many evaluations to return

        Returns
        -------
        tuple[EvaluationResult]
            the best evaluations from the AutoRAG experiment
        """
        evaluations = list(self.evaluations)
        sorted_evs = tuple(sorted(evaluations, key=lambda x: x.final_score, reverse=True))
        return sorted_evs[:k]

    @staticmethod
    def create_evaluation_results_json(
        evaluation_data: list[EvaluationData], evaluation_result: EvaluationResult
    ) -> list[dict[str, Any]]:
        """
        Create json made of evaluation results with proper data.

        Example file content:
        [
            {
                "question_id": "0",
                "answer": "<model's answer>",
                "answer_contexts": [
                    {"text": "<content1_text>", "document_id": "document_1.pdf"},
                    {"text": "<content2_text>", "document_id": "document_2.pdf"},
                ]
                "scores": {
                    "answer_correctness": 0.79,
                    "context_correctness": 0.65,
                }
            },
            {
                "question_id": "1",
                "answer": "<model's answer>",
                "answer_contexts": [
                    {"text": "<content3_text>", "document_id": "document_3.pdf"},
                    {"text": "<content4_text>", "document_id": "document_4.pdf"},
                ]
                "scores": {
                    "answer_correctness": 0.79,
                    "context_correctness": 0.65,
                }
            },
        ]

        Parameters
        ----------
        evaluation_data : list[EvaluationData]
            Evaluation data to be used for the json generation.

        evaluation_result : EvaluationResult
            Evaluation result to be used for the json generation.

        Returns
        -------
        list[dict[str, Any]]
            json-like object with evaluation results.
        """
        data = []
        for local_ev_data in evaluation_data:
            ret = {
                "question": local_ev_data.question,
                "correct_answers": local_ev_data.ground_truths,
                "answer": local_ev_data.answer,
                "answer_contexts": [
                    {"text": text, "document_id": doc_id}
                    for text, doc_id in zip(local_ev_data.contexts, local_ev_data.context_ids)
                ],
                "scores": {
                    key: evaluation_result.scores["question_scores"][key][local_ev_data.question_id]
                    for key in evaluation_result.scores["question_scores"]
                },
            }
            data.append(ret)

        return data



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
class SearchSpaceError(Exception):
    """Error representing problem with search space initialisation"""


class SearchSpaceValueError(ValueError):
    """Error representing incorrect value given for search space type"""



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import random
from dataclasses import dataclass
from typing import Any, Callable

from ai4rag import logger
from ai4rag.core.hpo.base_optimizer import BaseOptimizer, FailedIterationError, OptimizationError, OptimizerSettings
from ai4rag.search_space.src.search_space import SearchSpace

__all__ = ["RandomOptimizer", "RandomOptSettings", "FailedIterationError"]


@dataclass
class RandomOptSettings(OptimizerSettings):
    """Settings for random optimizer."""


class RandomOptimizer(BaseOptimizer):
    """Optimizer running random search on the given search space."""

    def __init__(
        self, objective_function: Callable[[dict], float], search_space: SearchSpace, settings: RandomOptSettings
    ):
        super().__init__(objective_function, search_space, settings)
        self._evaluated_combinations = []

    def search(self) -> dict[str, Any]:
        """
        Actual function performing hyperparameter optimization for the selected
        objective function.

        Returns
        -------
        dict[str, Any]
            The best set of parameters with achieved score.

        Raises
        ------
        OptimizationError
            When there were no successful evaluations for given constraints.
        """
        combinations = self._search_space.combinations
        random.shuffle(combinations)

        for idx in range(self.settings.max_evals):
            score = self._objective_function(combinations[idx])
            self._evaluated_combinations.append(combinations[idx] | {"score": score})

        successful_evaluations = [ev for ev in self._evaluated_combinations if ev["score"] is not None]

        if not successful_evaluations:
            raise OptimizationError("Number of evaluations has reached limit. All iterations have failed.")

        best_config_with_score = sorted(successful_evaluations, key=lambda d: d["score"])[-1]

        return best_config_with_score

    def _objective_function(self, params: dict) -> float | None:
        """
        Wrapper around the objective function provided to the optimizer.

        Parameters
        ----------
        params : dict
            A dictionary containing parameters of pattern to be evaluated.

        Returns
        -------
        float | None
            Optimization score achieved for single node evaluation.
            If None - iteration has ended up with a failed status.
        """

        try:
            logger.info("Evaluating objective function with parameters: %s", params)
            loss = self.objective_function(params)

        except FailedIterationError:
            # None is here to avoid penalization of iterations failing due to unknown reasons
            loss = None

        return loss



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from langchain_core.documents import Document
from llama_stack_client import LlamaStackClient
from llama_stack_client.types.vector_store import VectorStore

from ai4rag.rag.embedding.llama_stack import LSEmbeddingModel
from ai4rag.rag.vector_store.base_vector_store import BaseVectorStore


class LSVectorStore(BaseVectorStore):
    """LLamaStack client wrapper used for communication with vector store (single index/collection)."""

    _VALID_SEARCH_MODES = ("vector", "hybrid")
    _VALID_RANKER_STRATEGIES = ("rrf", "weighted", "normalized")

    def __init__(
        self,
        embedding_model: LSEmbeddingModel,
        client: LlamaStackClient,
        provider_id: str,
        reuse_collection_name: str | None = None,
        distance_metric: str | None = None,
    ):
        super().__init__(embedding_model, distance_metric, reuse_collection_name)
        self.client = client
        self._ls_vs = self._initialize_ls_vector_store(
            client=client,
            embedding_model=embedding_model,
            provider_id=provider_id,
            reuse_collection_name=reuse_collection_name,
        )
        self._collection_name = self._ls_vs.id

    @staticmethod
    def _initialize_ls_vector_store(
        client: LlamaStackClient, embedding_model: LSEmbeddingModel, provider_id: str, reuse_collection_name: str | None
    ) -> VectorStore:
        """
        Create or retrieve vector store instance via llama-stack.

        Parameters
        ----------
        client : LlamaStackClient
            Llama-stack client instance for communication with llama-stack.

        embedding_model : LSEmbeddingModel
            Wrapped llama-stack based embedding model with proper parameters.

        provider_id : str
            Provider id within the llama stack server.

        reuse_collection_name : str | None
            vector_store_id within llama-stack-server (usually collection name) to reuse (if already existing)

        Returns
        -------
        llama_stack_client.types.vector_store.VectorStore
            Instance for communication with llama-stack vector store.
        """

        if reuse_collection_name:
            _vs = client.vector_stores.retrieve(reuse_collection_name)
            return _vs

        # Handle both dict and LSEmbeddingParams for backward compatibility
        if isinstance(embedding_model.params, dict):
            embedding_dimension = embedding_model.params["embedding_dimension"]
        else:
            embedding_dimension = embedding_model.params.embedding_dimension

        _vs = client.vector_stores.create(
            extra_body={
                "provider_id": provider_id,
                "embedding_model": embedding_model.model_id,
                "embedding_dimension": embedding_dimension,
            }
        )

        return _vs

    @property
    def collection_name(self) -> str:
        return self._collection_name

    @staticmethod
    def _validate_search_params(
        search_mode: str,
        ranker_strategy: str | None,
        ranker_k: int | None,
        ranker_alpha: float | None,
    ) -> None:
        """Validate hybrid search parameter consistency."""
        if search_mode not in LSVectorStore._VALID_SEARCH_MODES:
            raise ValueError(
                f"Invalid search_mode '{search_mode}'. Must be one of {LSVectorStore._VALID_SEARCH_MODES}."
            )

        has_strategy = ranker_strategy is not None and ranker_strategy != ""
        has_k = ranker_k is not None and ranker_k > 0
        has_alpha = ranker_alpha is not None and ranker_alpha != 1

        if search_mode != "hybrid":
            if has_strategy:
                raise ValueError(
                    f"ranker_strategy='{ranker_strategy}' is only valid when search_mode='hybrid', "
                    f"but search_mode='{search_mode}'."
                )
            if has_k:
                raise ValueError(
                    f"ranker_k={ranker_k} is only valid when search_mode='hybrid', " f"but search_mode='{search_mode}'."
                )
            if has_alpha:
                raise ValueError(
                    f"ranker_alpha={ranker_alpha} is only valid when search_mode='hybrid', "
                    f"but search_mode='{search_mode}'."
                )
        else:
            if not has_strategy:
                raise ValueError("ranker_strategy must be set when search_mode='hybrid'.")
            if ranker_strategy not in LSVectorStore._VALID_RANKER_STRATEGIES:
                raise ValueError(
                    f"Invalid ranker_strategy='{ranker_strategy}'. "
                    f"Must be one of {LSVectorStore._VALID_RANKER_STRATEGIES}."
                )
            if has_k and ranker_strategy != "rrf":
                raise ValueError(
                    f"ranker_k={ranker_k} is only valid when ranker_strategy='rrf', "
                    f"but ranker_strategy='{ranker_strategy}'."
                )
            if has_alpha and ranker_strategy != "weighted":
                raise ValueError(
                    f"ranker_alpha={ranker_alpha} is only valid when ranker_strategy='weighted', "
                    f"but ranker_strategy='{ranker_strategy}'."
                )

    def search(
        self,
        query: str,
        k: int,
        include_scores: bool = False,
        search_mode: str = "vector",
        ranker_strategy: str | None = None,
        ranker_k: int | None = None,
        ranker_alpha: float | None = None,
        **kwargs,
    ) -> list[Document] | list[tuple[Document, float]]:
        """
        Search for the chunks relevant to the query.

        Parameters
        ----------
        query : str
            Question / query for which the similarity search will be executed.

        k : int
            Number of chunks to be returned as a result of similarity search

        include_scores : bool, default=False
            If True, similarity scores will be returned in the response

        search_mode : str, default="vector"
            Search mode: "vector" or "hybrid".

        ranker_strategy : str | None, default=None
            Ranking strategy for hybrid search: "rrf", "weighted", or "normalized".
            Empty string means no ranker (used for non-hybrid modes).

        ranker_k : int | None, default=None
            Parameter k for the ranking function. 0 means not set.

        ranker_alpha : float, default=None
            Alpha parameter for weighted ranking strategy. 1 means not set (vector-only sentinel).

        Returns
        -------
        list[Document] | list[tuple[Document, float]]
            List of chunks as Document instances with or without scores, depending on the input.
        """
        self._validate_search_params(search_mode, ranker_strategy, ranker_k, ranker_alpha)
        params = {
            "max_chunks": k,
            "mode": search_mode,
        }

        if search_mode == "hybrid" and ranker_strategy:
            params["reranker_type"] = ranker_strategy
            reranker_params = {}
            if ranker_strategy == "rrf" and ranker_k is not None and ranker_k > 0:
                reranker_params["impact_factor"] = ranker_k
            if ranker_strategy == "weighted" and ranker_alpha is not None and ranker_alpha != 1:
                reranker_params["alpha"] = ranker_alpha
            params["reranker_params"] = reranker_params

        resp = self.client.vector_io.query(query=query, vector_store_id=self._ls_vs.id, params=params)

        if include_scores:
            return [
                (Document(page_content=chunk.content, metadata=chunk.chunk_metadata.to_dict()), score)
                for chunk, score in zip(resp.chunks, resp.scores)
            ]

        return [Document(page_content=chunk.content, metadata=chunk.chunk_metadata.to_dict()) for chunk in resp.chunks]

    def add_documents(self, documents: list[Document]) -> None:
        """
        Add documents to the collection.

        Parameters
        ----------
        documents : Sequence[Document]
            Documents to add to the collection.
        """

        # Handle both dict and LSEmbeddingParams for backward compatibility
        if isinstance(self.embedding_model.params, dict):
            embedding_dimension = self.embedding_model.params["embedding_dimension"]
        else:
            embedding_dimension = self.embedding_model.params.embedding_dimension

        chunks = [
            {
                "content": doc.page_content,
                "chunk_metadata": doc.metadata,
                "chunk_id": doc.metadata["document_id"],
                "embedding_model": self.embedding_model.model_id,
                "embedding_dimension": embedding_dimension,
            }
            for doc in documents
        ]
        embeddings = self.embedding_model.embed_documents([doc.page_content for doc in documents])
        full_chunks = [chunk | {"embedding": embedding} for chunk, embedding in zip(chunks, embeddings)]
        self.client.vector_io.insert(
            vector_store_id=self._ls_vs.id,
            chunks=full_chunks,
        )

    def clean_collection(self):
        """Remove content of the collection and remove vector store instance."""
        self.client.vector_stores.delete(self._ls_vs.id)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod

from ..embedding.base_model import BaseEmbeddingModel
from ..foundation_models.base_model import BaseFoundationModel
from ..retrieval.retriever import Retriever
from ..vector_store.base_vector_store import BaseVectorStore


class RAGTemplateError(Exception):
    """Instance raised when error occurs in the RAG template."""


class BaseRAGTemplate(ABC):
    """
    Base abstract class for Retrieval-Augmented Generation (RAG) templates.

    This class defines the interface for RAG templates that combine embedding models,
    vector stores, retrievers, and foundation models to enable question-answering
    over custom document collections.

    A RAG template orchestrates the following workflow:
    1. Index building: Process and store documents in a vector store
    2. Retrieval: Find relevant documents for a given query
    3. Generation: Use a foundation model to generate answers based on retrieved context

    Parameters
    ----------
    foundation_model : BaseFoundationModel
        The foundation model (LLM) used to generate answers based on retrieved context.

    retriever : Retriever
        The retriever component responsible for finding relevant documents from the vector store.

    embedding_model : BaseEmbeddingModel | None, default=None
        The embedding model used to convert text into vector representations.

    vector_store : BaseVectorStore | None, default=None
        The vector store that maintains indexed documents and supports similarity search.

    Notes
    -----
    Subclasses must implement all abstract methods: build_index, generate, generate_stream.
    """

    def __init__(
        self,
        foundation_model: BaseFoundationModel,
        retriever: Retriever,
        vector_store: BaseVectorStore | None = None,
        embedding_model: BaseEmbeddingModel | None = None,
    ):
        self.embedding_model: BaseEmbeddingModel = embedding_model
        self.foundation_model: BaseFoundationModel = foundation_model
        self.retriever: Retriever = retriever
        self.vector_store: BaseVectorStore = vector_store

    @abstractmethod
    def build_index(
        self,
        *args,
        **kwargs,
    ):
        """Index building method."""

    @abstractmethod
    def generate(
        self,
        *args,
        **kwargs,
    ):
        """Template generation method."""

    @abstractmethod
    def generate_stream(
        self,
        *args,
        **kwargs,
    ):
        """Template generation stream method."""



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from string import Formatter
from typing import Literal, TypeVar

from ai4rag.search_space.src.model_props import (
    CONTEXT_TEXT_PLACEHOLDER,
    QUESTION_PLACEHOLDER,
    REFERENCE_DOCUMENTS_PLACEHOLDER,
)
from ai4rag.utils.validators import ConstraintsValidationError, OneOf, Validator

T = TypeVar("T")


class RAGPromptTemplateString(Validator[str]):
    """Validates RAG template string."""

    template_name: OneOf[Literal["context_template_text", "user_message_text"]] = OneOf(
        "context_template_text", "user_message_text"
    )

    def __init__(
        self,
        template_name: Literal["context_template_text", "user_message_text"],
    ) -> None:
        super().__init__()
        self.template_name = template_name

        self._required_placeholders: tuple[str, ...] = (
            (CONTEXT_TEXT_PLACEHOLDER,)
            if template_name == "context_template_text"
            else (QUESTION_PLACEHOLDER, REFERENCE_DOCUMENTS_PLACEHOLDER)
        )

    def validate(self, _: object, value: T) -> T:
        """
        Validates if user provided correct placeholders in given template text in respect to default placeholders.

        Parameters
        ----------
        template_str : str
            Prompt template with proper placeholders to be validated.

        template_name : Literal["context_template_text", "user_message_text"]
            Name of the template that will be validated. Used for required placeholders selection.

        Returns
        -------
        str
            Prompt template with filled placeholders.

        Raises
        ------
        ValueError
            When user provided less placeholders than expected.

            When user provided wrong placeholder name.
        """
        if not isinstance(value, str):
            raise TypeError(f"Expected {value!r} to be a str or None.")

        placeholders_count = 0

        for _, field_name, _, _ in Formatter().parse(value):
            if field_name is None:
                # when there is text NOT followed by a placeholder template
                continue
            if field_name not in self._required_placeholders:
                raise ConstraintsValidationError(
                    f"Custom {field_name.split('_')[0]} template text got unexpected placeholder `{field_name}`, "
                    f"valid placeholders are `{self._required_placeholders}`."
                )

            placeholders_count += 1

        if placeholders_count != len(self._required_placeholders):
            raise ConstraintsValidationError(
                f"Incorrect number of placeholders required for: '{value.split('_')[0]}' template text, "
                f"expected {len(self._required_placeholders)} but got {placeholders_count}."
            )

        return value


def _validate_prompt_templates_placeholders(
    template_str: str,
    template_name: Literal["context_template_text", "user_message_text"],
) -> str:
    """
    Validates if user provided correct placeholders in given template text in respect to default placeholders.

    Parameters
    ----------
    template_str : str
        Prompt template with proper placeholders to be validated.

    template_name : Literal["context_template_text", "user_message_text"]
        Name of the template that will be validated. Used for required placeholders selection.

    Returns
    -------
    str
        Prompt template with filled placeholders.

    Raises
    ------
    ValueError
        When user provided less placeholders than expected.

        When user provided wrong placeholder name.
    """
    if template_name == "context_template_text":
        required_placeholders = (CONTEXT_TEXT_PLACEHOLDER,)
    elif template_name == "user_message_text":
        required_placeholders = (QUESTION_PLACEHOLDER, REFERENCE_DOCUMENTS_PLACEHOLDER)
    else:
        raise ValueError(f"Cannot validate presence of expected template placeholders on field: {template_name}")

    placeholders_count = 0

    for _, field_name, _, _ in Formatter().parse(template_str):
        if field_name is None:
            # when there is text NOT followed by a placeholder template
            continue
        if field_name not in required_placeholders:
            raise ValueError(
                f"Custom {field_name.split('_')[0]} template text got unexpected placeholder `{field_name}`, "
                f"valid placeholders are `{required_placeholders}`."
            )

        placeholders_count += 1

    if placeholders_count != len(required_placeholders):
        raise ValueError(
            f"Incorrect number of placeholders required for {template_name.split('_')[0]} template text, "
            f"expected {len(required_placeholders)} but got {placeholders_count}."
        )
    return template_str



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any

__all__ = [
    "ConstantMeta",
    "AI4RAGParamNames",
    "GenerationConstants",
    "ChunkingConstraints",
    "ExperimentStep",
    "RetrievalConstraints",
    "SearchSpaceValidationErrors",
    "DEFAULT_WORD_TO_TOKEN_RATIO",
    "ChatGenerationConstants",
    "DefaultVectorStoreFieldNames",
]

DEFAULT_WORD_TO_TOKEN_RATIO = 1.5


class ConstantMeta(type):
    """Metaclass for all instance classes that we desire to have constant attributes."""

    def __new__(mcs, name, bases, class_dict):
        _constant_attributes = [
            val
            for key, val in class_dict.items()
            if not key.startswith("__") and not callable(val) and type(val) in (str, int, float)
        ]
        class_dict["_constant_attributes"] = _constant_attributes

        new_class = super().__new__(mcs, name, bases, class_dict)

        return new_class

    def __setattr__(cls, name, value) -> None:
        raise AttributeError(f"Cannot modify attribute '{name}' after class creation.")

    def __iter__(cls) -> Any:
        yield from cls._constant_attributes

    def __contains__(cls, value: Any) -> bool:
        return value in cls._constant_attributes

    def validate(cls, value: Any) -> Any:
        """Validates if given value exists in defined constants.

        Parameters
        ----------
        value : Any
            Value to search in defined constants.

        Returns
        -------
        Any
            Returns provided value if valid.

        Raises
        ------
        ValueError
            When value doesn't exists in declared constants.
        """
        if value not in cls._constant_attributes:
            raise ValueError(f"Value {value} not found in defined constants.")
        return value


class AI4RAGParamNames(metaclass=ConstantMeta):
    """Parameter's names used in the experiment."""

    CHUNKING = "chunking"
    CHUNKING_METHOD = "chunking_method"
    CHUNK_SIZE = "chunk_size"
    CHUNK_OVERLAP = "chunk_overlap"
    EMBEDDING_MODEL = "embedding_model"
    DISTANCE_METRIC = "distance_metric"
    FOUNDATION_MODEL = "foundation_model"
    TRUNCATE_STRATEGY = "truncate_strategy"
    INPUT_SIZE = "input_size"
    RETRIEVAL = "retrieval"
    RETRIEVAL_METHOD = "retrieval_method"
    WINDOW_SIZE = "window_size"
    NUMBER_OF_CHUNKS = "number_of_chunks"
    SEARCH_MODE = "search_mode"
    RANKER_STRATEGY = "ranker_strategy"
    RANKER_K = "ranker_k"
    RANKER_ALPHA = "ranker_alpha"
    GENERATION = "generation"


class ExperimentStep(metaclass=ConstantMeta):
    """Steps occurring in the experiment engine."""

    MODEL_SELECTION = "model selection"
    OPTIMIZATION = "optimization"
    CHUNKING = "chunking"
    EMBEDDING = "embedding"
    GENERATION = "generation"
    EVALUATION = "evaluation"


class GenerationConstants(metaclass=ConstantMeta):
    """Constants used for setting the generation (inference) parameters."""

    MAX_NEW_TOKENS = 1000
    MIN_NEW_TOKENS = 1
    DECODING_METHOD = "greedy"


class ChatGenerationConstants(metaclass=ConstantMeta):
    """Constants used for setting the generation (inference) parameters for chat models only."""

    MAX_COMPLETION_TOKENS = 2048
    TEMPERATURE = 0.2


class ChunkingConstraints(metaclass=ConstantMeta):
    """Constants used to define chunking constraints on what below parameters can be."""

    MIN_CHUNK_SIZE = 128
    MAX_CHUNK_SIZE = 2048
    MIN_CHUNK_OVERLAP = 64
    MAX_CHUNK_OVERLAP = 512
    METHODS = ["recursive", "semantic"]


class RetrievalConstraints(metaclass=ConstantMeta):
    """Constants used to define the permissible values for retrieval constraints."""

    MIN_NUMBER_OF_RETRIEVED_CHUNKS = 1
    MAX_NUMBER_OF_RETRIEVED_CHUNKS = 10
    METHODS = ["window", "simple"]
    MIN_WINDOW_SIZE = 0
    MAX_WINDOW_SIZE = 4
    SEARCH_MODES = ["vector", "hybrid"]
    RANKER_STRATEGIES = ["rrf", "weighted", "normalized"]
    MIN_RANKER_K = 1
    MAX_RANKER_K = 100


class DefaultVectorStoreFieldNames(metaclass=ConstantMeta):
    """Constants used as field names during index building for Milvus and Elasticsearch"""

    CHUNK_SEQUENCE_NUMBER_FIELD = "sequence_number"
    DENSE_EMBEDDINGS_FIELD = "vector"
    SPARSE_EMBEDDINGS_FIELD = "sparse_embeddings"
    MILVUS_TEXT_FIELD = "text"
    ELASTICSEARCH_TEXT_FIELD = "text_field"
    METADATA_DOCUMENT_NAME_FIELD = "document_id"


class SearchSpaceValidationErrors(metaclass=ConstantMeta):
    """Constants used to define payload validation errors."""

    UNEXPECTED_KEYWORD_ARGUMENT = "unexpected_keyword_argument"
    LITERAL_ERROR = "literal_error"
    TOO_SHORT = "too_short"
    TOO_LONG = "too_long"
    LESS_THAN_EQUAL = "less_than_equal"
    GREATER_THAN_EQUAL = "greater_than_equal"
    INT_FROM_FLOAT = "int_from_float"
    INT_PARSING = "int_parsing"
    INT_TYPE = "int_type"
    LIST_TYPE = "list_type"



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from typing import Any, Generic, Sequence, TypeVar

__all__ = [
    "BaseChunker",
]

ChunkT = TypeVar("ChunkT")


class BaseChunker(ABC, Generic[ChunkT]):
    """
    Responsible for handling splitting document operations
    in the RAG application.
    """

    @abstractmethod
    def split_documents(self, documents: Sequence[ChunkT]) -> list[ChunkT]:
        """
        Split series of documents into smaller parts based on
        the provided chunker settings.

        Parameters
        ----------
        documents : Sequence[ChunkType]
            Sequence of elements that contain context in a text format.

        Returns
        -------
        list[ChunkType]
            List of documents split into smaller ones, having less content.
        """

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Return dictionary that can be used to recreate an instance of the BaseChunker."""

    @classmethod
    @abstractmethod
    def from_dict(cls, d: dict[str, Any]) -> "BaseChunker":
        """Create an instance from the dictionary."""



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Iterator

import pandas as pd


class BenchmarkDataValueError(ValueError):
    """Error representing incorrect value given in the benchmark dataset"""


class BenchmarkData:
    """
    Class representing benchmarking dataset given to the AI4RAGExperiment
    introducing user-friendly, specified interface.

    Parameters
    ----------
    benchmark_data : pandas.DataFrame
        Benchmark Data given as DataFrame.It should be tabular data that contains
        questions, answers and documents_ids columns.

    Attributes
    ----------
    questions : list[str]
        Validated questions from the benchmark dataset.

    correct_answers : list[str]
        Validated answers from the benchmark dataset.

    document_ids : list[str]
        Validated IDs of documents with correct context for given answers.

    Raises
    ------
    BenchmarkValueError
        Raised when any of the arguments in the dataset is considered invalid.
    """

    QUESTION = "question"
    CORRECT_ANSWERS = "correct_answers"
    DOC_IDS = "correct_answer_document_ids"

    def __init__(self, benchmark_data: pd.DataFrame):
        if len(benchmark_data) == 0:
            raise BenchmarkDataValueError("There are no records in the benchmark data.")
        self._benchmark_data = benchmark_data

        self.questions: list[str] = list(self._benchmark_data[self.QUESTION])
        self.correct_answers: list[list[str]] = list(self._benchmark_data[self.CORRECT_ANSWERS])
        self.document_ids: list[list[str]] = list(self._benchmark_data[self.DOC_IDS])
        self._questions_ids = [f"q{idx}" for idx in range(len(self.questions))]

    def __iter__(self) -> Iterator[tuple[str, list[str], list[str] | None]]:
        for q, a, id_ in zip(self.questions, self.correct_answers, self.document_ids):
            yield q, a, id_

    def __len__(self) -> int:
        return len(self.questions)

    def __getitem__(self, idx: int) -> tuple[str, list[str], list[str] | None]:
        return self.questions[idx], self.correct_answers[idx], self.document_ids[idx]

    def get_random_sample(self, n_records: int = 10, random_seed: int = 17) -> "BenchmarkData":
        """
        Create sample of the original BenchmarkData. If number of desired records
        is bigger than actual size of the data, create new instance based on
        all samples.

        Parameters
        ----------
        n_records : int, default=10
            Number of records to be included in the newly created instance.

        random_seed : int, default=17
            Seed to make data sampling deterministic.

        Returns
        -------
        BenchmarkData
            New instance of BenchmarkData.
        """
        if n_records <= 0:
            raise BenchmarkDataValueError("Cannot make empty sample. Select 'n_records' to be an int greater than 0.")
        if n_records > len(self):
            sample = self._benchmark_data.copy()
        else:
            sample = self._benchmark_data.sample(n=n_records, random_state=random_seed)
        return self.__class__(benchmark_data=sample)

    @property
    def questions(self) -> list[str]:
        """Get all questions from benchmark data."""
        return self._questions

    @questions.setter
    def questions(self, val: list[str]) -> None:
        """Validate whether questions is a list of not empty strings"""
        _validate_list_of_strings(val, self.QUESTION)

        self._questions = val

    @property
    def correct_answers(self) -> list[list[str]]:
        """get all answers from benchmark data."""
        return self._correct_answers

    @correct_answers.setter
    def correct_answers(self, val: list[list[str]]) -> None:
        """Validate whether each element is a list of not empty strings"""
        for el in val:
            _validate_list_of_strings(el, self.CORRECT_ANSWERS)
        self._correct_answers = val

    @property
    def document_ids(self) -> list[list[str]]:
        """Get all document ids from benchmark data."""
        return self._document_ids

    @document_ids.setter
    def document_ids(self, val: list[list[str]] | None) -> None:
        """Validate whether each element is a list of not empty strings"""
        if val is None:
            self._document_ids = val
        else:
            for el in val:
                _validate_list_of_strings(el, self.DOC_IDS)
            self._document_ids = val

    @property
    def questions_ids(self) -> list[str]:
        """Get all questions ids from benchmark data."""
        return self._questions_ids


def _validate_list_of_strings(elements: list[str], key: str) -> None:
    """
    Validate whether list of values is actually list of not-empty strings.

    Parameters
    ----------
    elements : list[str]
        List to be validated

    key : str
        What attribute we are validating. It is used to create proper message

    Raises
    ------
    BenchmarkDataValueError
        When some element is invalid
    """

    for element in elements:
        if not isinstance(element, str) or not element:
            raise BenchmarkDataValueError(f"Incorrect '{key}' value: '{element}'.")



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any

from llama_stack_client import LlamaStackClient
from pydantic import TypeAdapter, ValidationError

from ai4rag import logger
from ai4rag.rag.embedding.llama_stack import LSEmbeddingModel
from ai4rag.rag.foundation_models.llama_stack import LSFoundationModel
from ai4rag.search_space.prepare.input_payload_types import AI4RAGConstraints
from ai4rag.search_space.prepare.llama_stack_utils import (
    _are_provided_models_available,
    _get_default_llama_stack_models,
)
from ai4rag.search_space.prepare.validation_error_decoder import validation_error_decoder
from ai4rag.search_space.src.exceptions import SearchSpaceValueError
from ai4rag.search_space.src.parameter import Parameter
from ai4rag.search_space.src.search_space import AI4RAGSearchSpace

__all__ = ["prepare_search_space_with_llama_stack"]


def prepare_search_space_with_llama_stack(
    payload: dict[str, Any],
    client: LlamaStackClient,
    vector_store_type: str = "ls_milvus",
) -> AI4RAGSearchSpace:
    """
    Prepare AutoRAGSearchSpace.

    Parameters
    ----------
    payload : dict[str, Any]
        A mapping between parameter name and its associated values.

    client : LlamaStackClient
        Client instance for listing and validating available models.

    vector_store_type : str, default="ls_milvus"
        Type of vector store. When "chroma", hybrid search parameters are excluded
        from the default search space since ChromaDB does not support hybrid search.

    Returns
    -------
    AI4RAGSearchSpace
        A valid AI4RAGSearchSpace used in RAG optimization process.

    Raises
    ------
    SearchSpaceValueError
        Raised when payload contains non-recognized parameter name.
    """
    logger.info("Preparing search space based on provided constraints: %s.", payload)

    payload_model = TypeAdapter(AI4RAGConstraints)

    try:
        validated_payload = payload_model.validate_python(payload)
    except ValidationError as ve:
        # we want to catch only the first error
        validation_error_decoder(ve.errors()[0])

    if isinstance(client, LlamaStackClient):
        models = _get_default_llama_stack_models(client)
        default_foundation_models = models["foundation_models"]
        default_embedding_models = models["embedding_models"]
    else:
        raise SearchSpaceValueError(f"Unrecognized client type: '{client.__class__.__name__}'")

    if validated_payload.foundation_models:
        _are_provided_models_available(validated_payload.foundation_models, default_foundation_models)
    if validated_payload.embedding_models:
        _are_provided_models_available(validated_payload.embedding_models, default_embedding_models)

    # Transform user models into llama-stack based models
    if validated_payload.foundation_models is not None:
        fms_param = Parameter(
            name="foundation_model",
            values=[
                LSFoundationModel(
                    model_id=fm.model_id,
                    client=client,
                )
                for fm in validated_payload.foundation_models
            ],
        )
    else:
        fms_param = Parameter(
            name="foundation_model",
            values=default_foundation_models,
        )

    if validated_payload.embedding_models is not None:
        embedding_models_values = []
        for em in validated_payload.embedding_models:
            matched_model = next(filter(lambda x, _id=em.model_id: x.model_id == _id, default_embedding_models), None)
            if matched_model is None:
                raise SearchSpaceValueError(f"Embedding model '{em.model_id}' not found among available models.")
            embedding_models_values.append(
                LSEmbeddingModel(model_id=em.model_id, client=client, params=matched_model.params)
            )
        ems_param = Parameter(
            name="embedding_model",
            values=embedding_models_values,
        )
    else:
        ems_param = Parameter(
            name="embedding_model",
            values=default_embedding_models,
        )

    return AI4RAGSearchSpace(
        params=[fms_param, ems_param],
        vector_store_type=vector_store_type,
    )
